import itertools
import logging
from collections import defaultdict, deque
from operator import mul
from sqlalchemy.dialects import postgresql

from raco import algebra, expression, rules
from raco.catalog import Catalog
from raco.language import Language, Algebra
from raco.language.sql.catalog import SQLCatalog
from raco.expression import UnnamedAttributeRef
from raco.expression.aggregate import (rebase_local_aggregate_output,
                                       rebase_finalizer)
from raco.expression.statevar import *
from raco.datastructure.UnionFind import UnionFind
from raco import types

LOGGER = logging.getLogger(__name__)


def scheme_to_schema(s):
    if s:
        names, descrs = zip(*s.asdict.items())
        names = ["%s" % n for n in names]
        types_ = [r[1] for r in descrs]
    else:
        names = []
        types_ = []
    return {"columnTypes": types_, "columnNames": names}


def compile_expr(op, child_scheme, state_scheme):
    ####
    # Put special handling at the top!
    ####
    if isinstance(op, expression.NumericLiteral):
        if type(op.value) == int:
            myria_type = types.LONG_TYPE
        elif type(op.value) == float:
            myria_type = types.DOUBLE_TYPE
        else:
            raise NotImplementedError("Compiling NumericLiteral {} of type {}"
                                      .format(op, type(op.value)))

        return {
            'type': 'CONSTANT',
            'value': str(op.value),
            'valueType': myria_type
        }
    elif isinstance(op, expression.StringLiteral):
        return {
            'type': 'CONSTANT',
            'value': str(op.value),
            'valueType': types.STRING_TYPE
        }
    elif isinstance(op, expression.BooleanLiteral):
        return {
            'type': 'CONSTANT',
            'value': bool(op.value),
            'valueType': 'BOOLEAN_TYPE'
        }
    elif isinstance(op, expression.StateRef):
        return {
            'type': 'STATE',
            'columnIdx': op.get_position(child_scheme, state_scheme)
        }
    elif isinstance(op, expression.AttributeRef):
        return {
            'type': 'VARIABLE',
            'columnIdx': op.get_position(child_scheme, state_scheme)
        }
    elif isinstance(op, expression.Case):
        # Convert n-ary case statements to binary, as expected by Myria
        op = op.to_binary()
        assert len(op.when_tuples) == 1

        if_expr = compile_expr(op.when_tuples[0][0], child_scheme,
                               state_scheme)
        then_expr = compile_expr(op.when_tuples[0][1], child_scheme,
                                 state_scheme)
        else_expr = compile_expr(op.else_expr, child_scheme, state_scheme)

        return {
            'type': 'CONDITION',
            'children': [if_expr, then_expr, else_expr]
        }
    elif isinstance(op, expression.CAST):
        return {
            'type': 'CAST',
            'left': compile_expr(op.input, child_scheme, state_scheme),
            'right': {
                'type': 'TYPE',
                'outputType': op._type
            }
        }

    ####
    # Everything below here is compiled automatically
    ####
    elif isinstance(op, expression.UnaryOperator):
        return {
            'type': op.opname(),
            'operand': compile_expr(op.input, child_scheme, state_scheme)
        }
    elif isinstance(op, expression.BinaryOperator):
        return {
            'type': op.opname(),
            'left': compile_expr(op.left, child_scheme, state_scheme),
            'right': compile_expr(op.right, child_scheme, state_scheme)
        }
    elif isinstance(op, expression.ZeroaryOperator):
        return {
            'type': op.opname(),
        }
    elif isinstance(op, expression.NaryOperator):
        children = []
        for operand in op.operands:
            children.append(compile_expr(operand, child_scheme, state_scheme))
        return {
            'type': op.opname(),
            'children': children
        }
    raise NotImplementedError("Compiling expr of class %s" % op.__class__)


def compile_mapping(expr, child_scheme, state_scheme):
    output_name, root_op = expr
    return {
        'outputName': output_name,
        'rootExpressionOperator': compile_expr(root_op,
                                               child_scheme,
                                               state_scheme)
    }


class MyriaLanguage(Language):
    reusescans = False


class MyriaOperator(object):
    language = MyriaLanguage


def relation_key_to_json(relation_key):
    return {"userName": relation_key.user,
            "programName": relation_key.program,
            "relationName": relation_key.relation}


class MyriaScan(algebra.Scan, MyriaOperator):
    def compileme(self):
        return {
            "opType": "TableScan",
            "relationKey": relation_key_to_json(self.relation_key),
        }


class MyriaScanTemp(algebra.ScanTemp, MyriaOperator):
    def compileme(self):
        return {
            "opType": "TempTableScan",
            "table": self.name,
        }


class MyriaUnionAll(algebra.UnionAll, MyriaOperator):
    def compileme(self, leftid, rightid):
        return {
            "opType": "UnionAll",
            "argChildren": [leftid, rightid]
        }


class MyriaDifference(algebra.Difference, MyriaOperator):
    def compileme(self, leftid, rightid):
        return {
            "opType": "Difference",
            "argChild1": leftid,
            "argChild2": rightid,
        }


class MyriaSingleton(algebra.SingletonRelation, MyriaOperator):
    def compileme(self):
        return {
            "opType": "Singleton",
        }


class MyriaEmptyRelation(algebra.EmptyRelation, MyriaOperator):
    def compileme(self):
        return {
            "opType": "Empty",
            'schema': scheme_to_schema(self.scheme())
        }


class MyriaSelect(algebra.Select, MyriaOperator):
    def compileme(self, inputid):
        pred = compile_expr(self.condition, self.scheme(), None)
        return {
            "opType": "Filter",
            "argChild": inputid,
            "argPredicate": {
                "rootExpressionOperator": pred
            }
        }


class MyriaCrossProduct(algebra.CrossProduct, MyriaOperator):
    def compileme(self, leftid, rightid):
        column_names = [name for (name, _) in self.scheme()]
        allleft = [i.position for i in self.left.scheme().ascolumnlist()]
        allright = [i.position for i in self.right.scheme().ascolumnlist()]
        return {
            "opType": "SymmetricHashJoin",
            "argColumnNames": column_names,
            "argChild1": leftid,
            "argChild2": rightid,
            "argColumns1": [],
            "argColumns2": [],
            "argSelect1": allleft,
            "argSelect2": allright
        }


class MyriaStore(algebra.Store, MyriaOperator):
    def compileme(self, inputid):
        return {
            "opType": "DbInsert",
            "relationKey": relation_key_to_json(self.relation_key),
            "argOverwriteTable": True,
            "argChild": inputid,
        }


class MyriaStoreTemp(algebra.StoreTemp, MyriaOperator):
    def compileme(self, inputid):
        return {
            "opType": "TempInsert",
            "table": self.name,
            "argOverwriteTable": True,
            "argChild": inputid,
        }


class MyriaAppendTemp(algebra.AppendTemp, MyriaOperator):
    def compileme(self, inputid):
        return {
            "opType": "TempInsert",
            "table": self.name,
            "argOverwriteTable": False,
            "argChild": inputid,
        }


def convertcondition(condition, left_len, combined_scheme):
    """Convert an equijoin condition to a pair of column lists."""

    if isinstance(condition, expression.AND):
        leftcols1, rightcols1 = convertcondition(condition.left,
                                                 left_len,
                                                 combined_scheme)
        leftcols2, rightcols2 = convertcondition(condition.right,
                                                 left_len,
                                                 combined_scheme)
        return leftcols1 + leftcols2, rightcols1 + rightcols2

    if isinstance(condition, expression.EQ):
        leftpos = condition.left.get_position(combined_scheme)
        rightpos = condition.right.get_position(combined_scheme)
        leftcol = min(leftpos, rightpos)
        rightcol = max(leftpos, rightpos)
        assert rightcol >= left_len
        return [leftcol], [rightcol - left_len]

    raise NotImplementedError("Myria only supports EquiJoins, not %s" % condition)  # noqa


def convert_nary_conditions(conditions, schemes):
    """Convert an NaryJoin map from global column index to local"""
    attr_map = {}   # map of global attribute to local column index
    count = 0
    for i, scheme in enumerate(schemes):
        for j, attr in enumerate(scheme.ascolumnlist()):
            attr_map[count] = [i, j]
            count += 1
    new_conditions = []   # arrays of [child_index, column_index]
    for join_cond in conditions:
        new_join_cond = []
        for attr in join_cond:
            new_join_cond.append(attr_map[attr.position])
        new_conditions.append(new_join_cond)
    return new_conditions


class MyriaSymmetricHashJoin(algebra.ProjectingJoin, MyriaOperator):
    def compileme(self, leftid, rightid):
        """Compile the operator to a sequence of json operators"""

        left_len = len(self.left.scheme())
        combined = self.left.scheme() + self.right.scheme()
        leftcols, rightcols = convertcondition(self.condition,
                                               left_len,
                                               combined)

        if self.output_columns is None:
            self.output_columns = self.scheme().ascolumnlist()
        column_names = [name for (name, _) in self.scheme()]
        pos = [i.get_position(combined) for i in self.output_columns]
        side = [p >= left_len for p in pos]
        assert sorted(side) == side, \
            "MyriaSymmetricHashJoin always emits left columns first"
        allleft = [i for i in pos if i < left_len]
        allright = [i - left_len for i in pos if i >= left_len]

        join = {
            "opType": "SymmetricHashJoin",
            "argColumnNames": column_names,
            "argChild1": "%s" % leftid,
            "argColumns1": leftcols,
            "argChild2": "%s" % rightid,
            "argColumns2": rightcols,
            "argSelect1": allleft,
            "argSelect2": allright
        }

        return join


class MyriaLeapFrogJoin(algebra.NaryJoin, MyriaOperator):

    def compileme(self, *args):
        def convert_join_cond(pos_to_rel_col, cond, scheme):
            join_col_pos = [c.get_position(scheme) for c in cond]
            return [pos_to_rel_col[p] for p in join_col_pos]
        # map a output column to its origin
        rel_of_pos = {}     # pos => [rel_idx, field_idx]
        schemes = [c.scheme().ascolumnlist() for c in self.children()]
        pos = 0
        combined = []
        for rel_idx, scheme in enumerate(schemes):
            combined.extend(scheme)
            for field_idx in xrange(len(scheme)):
                rel_of_pos[pos] = [rel_idx, field_idx]
                pos += 1
        # build column names
        if self.output_columns is None:
            self.output_columns = self.scheme().ascolumnlist()
        column_names = [name for (name, _) in self.scheme()]
        # get rel_idx and field_idx of select columns
        out_pos_list = [
            i.get_position(combined) for i in list(self.output_columns)]
        output_fields = [rel_of_pos[p] for p in out_pos_list]
        join_fields = [
            convert_join_cond(rel_of_pos, cond, combined)
            for cond in self.conditions]
        return {
            "opType": "LeapFrogJoin",
            "joinFieldMapping": join_fields,
            "argColumnNames": column_names,
            "outputFieldMapping": output_fields,
            "argChildren": args
        }


class MyriaGroupBy(algebra.GroupBy, MyriaOperator):
    @staticmethod
    def agg_mapping(agg_expr):
        """Maps a BuiltinAggregateExpression to a Myria string constant
        representing the corresponding aggregate operation."""
        if isinstance(agg_expr, expression.MAX):
            return "MAX"
        elif isinstance(agg_expr, expression.MIN):
            return "MIN"
        elif isinstance(agg_expr, expression.SUM):
            return "SUM"
        elif isinstance(agg_expr, expression.AVG):
            return "AVG"
        elif isinstance(agg_expr, expression.STDEV):
            return "STDEV"
        raise NotImplementedError("MyriaGroupBy.agg_mapping({})".format(
            type(agg_expr)))

    @staticmethod
    def compile_builtin_agg(agg, child_scheme):
        assert isinstance(agg, expression.BuiltinAggregateExpression)
        if isinstance(agg, expression.COUNTALL):
            return {"type": "CountAll"}

        assert isinstance(agg, expression.UnaryOperator)
        column = expression.toUnnamed(agg.input, child_scheme).position
        return {"type": "SingleColumn",
                "aggOps": [MyriaGroupBy.agg_mapping(agg)],
                "column": column}

    def compileme(self, inputid):
        child_scheme = self.input.scheme()
        group_fields = [expression.toUnnamed(ref, child_scheme)
                        for ref in self.grouping_list]

        built_ins = [agg_expr for agg_expr in self.aggregate_list
                     if isinstance(agg_expr,
                                   expression.BuiltinAggregateExpression)]

        aggregators = [MyriaGroupBy.compile_builtin_agg(agg_expr, child_scheme)
                       for agg_expr in built_ins]

        assert all(aggregators[i] != aggregators[j]
                   for i in range(len(aggregators))
                   for j in range(len(aggregators))
                   if i < j)

        udas = [agg_expr for agg_expr in self.aggregate_list
                if isinstance(agg_expr, expression.UdaAggregateExpression)]
        assert len(udas) + len(built_ins) == len(self.aggregate_list)

        if udas:
            inits = [compile_mapping(e, None, None) for e in self.inits]
            updates = [compile_mapping(e, child_scheme, self.state_scheme)
                       for e in self.updaters]
            emitters = [compile_mapping(("uda{i}".format(i=i),
                                         e.input),
                                        None, self.state_scheme)
                        for i, e in enumerate(udas)]
            aggregators.append({
                "type": "UserDefined",
                "initializers": inits,
                "updaters": updates,
                "emitters": emitters
            })

        ret = {
            "argChild": inputid,
            "aggregators": aggregators,
        }

        num_fields = len(self.grouping_list)
        if num_fields == 0:
            ret["opType"] = "Aggregate"
        elif num_fields == 1:
            ret["opType"] = "SingleGroupByAggregate"
            ret["argGroupField"] = group_fields[0].position
        else:
            ret["opType"] = "MultiGroupByAggregate"
            ret["argGroupFields"] = [field.position for field in group_fields]
        return ret


class MyriaInMemoryOrderBy(algebra.OrderBy, MyriaOperator):

    def compileme(self, inputsym):
        return {
            "opType": "InMemoryOrderBy",
            "argChild": inputsym,
            "argSortColumns": self.sort_columns,
            "argAscending": self.ascending
        }


class MyriaShuffle(algebra.Shuffle, MyriaOperator):
    """Represents a simple shuffle operator"""

    def compileme(self, inputid):
        raise NotImplementedError('shouldn''t ever get here, should be turned into SP-SC pair')  # noqa


class MyriaCollect(algebra.Collect, MyriaOperator):
    """Represents a simple collect operator"""

    def compileme(self, inputid):
        raise NotImplementedError('shouldn''t ever get here, should be turned into CP-CC pair')  # noqa


class MyriaDupElim(algebra.Distinct, MyriaOperator):
    """Represents duplicate elimination"""

    def compileme(self, inputid):
        return {
            "opType": "DupElim",
            "argChild": inputid,
        }


class MyriaApply(algebra.Apply, MyriaOperator):
    """Represents a simple apply operator"""

    def compileme(self, inputid):
        child_scheme = self.input.scheme()
        emitters = [compile_mapping(x, child_scheme, None)
                    for x in self.emitters]
        return {
            'opType': 'Apply',
            'argChild': inputid,
            'emitExpressions': emitters
        }


class MyriaStatefulApply(algebra.StatefulApply, MyriaOperator):
    """Represents a stateful apply operator"""

    def compileme(self, inputid):
        child_scheme = self.input.scheme()
        state_scheme = self.state_scheme
        comp_map = lambda x: compile_mapping(x, child_scheme, state_scheme)
        emitters = [comp_map(x) for x in self.emitters]
        inits = [comp_map(x) for x in self.inits]
        updaters = [comp_map(x) for x in self.updaters]
        return {
            'opType': 'StatefulApply',
            'argChild': inputid,
            'emitExpressions': emitters,
            'initializerExpressions': inits,
            'updaterExpressions': updaters
        }


class MyriaBroadcastProducer(algebra.UnaryOperator, MyriaOperator):
    """A Myria BroadcastProducer"""

    def __init__(self, input):
        algebra.UnaryOperator.__init__(self, input)

    def num_tuples(self):
        return self.input.num_tuples()

    def shortStr(self):
        return "%s" % self.opname()

    def compileme(self, inputid):
        return {
            "opType": "BroadcastProducer",
            "argChild": inputid,
        }


class MyriaBroadcastConsumer(algebra.UnaryOperator, MyriaOperator):
    """A Myria BroadcastConsumer"""

    def __init__(self, input):
        algebra.UnaryOperator.__init__(self, input)

    def num_tuples(self):
        return self.input.num_tuples()

    def shortStr(self):
        return "%s" % self.opname()

    def compileme(self, inputid):
        return {
            'opType': 'BroadcastConsumer',
            'argOperatorId': inputid
        }


class MyriaShuffleProducer(algebra.UnaryOperator, MyriaOperator):
    """A Myria ShuffleProducer"""

    def __init__(self, input, hash_columns):
        algebra.UnaryOperator.__init__(self, input)
        self.hash_columns = hash_columns

    def shortStr(self):
        hash_string = ','.join([str(x) for x in self.hash_columns])
        return "%s(h(%s))" % (self.opname(), hash_string)

    def __repr__(self):
        return "{op}({inp!r}, {hc!r})".format(op=self.opname(), inp=self.input,
                                              hc=self.hash_columns)

    def num_tuples(self):
        return self.input.num_tuples()

    def compileme(self, inputid):
        if len(self.hash_columns) == 1:
            pf = {
                "type": "SingleFieldHash",
                "index": self.hash_columns[0].position
            }
        else:
            pf = {
                "type": "MultiFieldHash",
                "indexes": [x.position for x in self.hash_columns]
            }

        return {
            "opType": "ShuffleProducer",
            "argChild": inputid,
            "argPf": pf
        }


class MyriaShuffleConsumer(algebra.UnaryOperator, MyriaOperator):
    """A Myria ShuffleConsumer"""

    def __init__(self, input):
        algebra.UnaryOperator.__init__(self, input)

    def num_tuples(self):
        return self.input.num_tuples()

    def shortStr(self):
        return "%s" % self.opname()

    def compileme(self, inputid):
        return {
            'opType': 'ShuffleConsumer',
            'argOperatorId': inputid
        }


class MyriaCollectProducer(algebra.UnaryOperator, MyriaOperator):
    """A Myria CollectProducer"""

    def __init__(self, input, server):
        algebra.UnaryOperator.__init__(self, input)
        self.server = server

    def num_tuples(self):
        return self.input.num_tuples()

    def shortStr(self):
        return "%s(@%s)" % (self.opname(), self.server)

    def compileme(self, inputid):
        return {
            "opType": "CollectProducer",
            "argChild": inputid,
        }

    def __repr__(self):
        return "{op}({inp!r}, {svr!r})".format(op=self.opname(),
                                               inp=self.input,
                                               svr=self.server)


class MyriaCollectConsumer(algebra.UnaryOperator, MyriaOperator):
    """A Myria CollectConsumer"""

    def __init__(self, input):
        algebra.UnaryOperator.__init__(self, input)

    def num_tuples(self):
        return self.input.num_tuples()

    def shortStr(self):
        return "%s" % self.opname()

    def compileme(self, inputid):
        return {
            'opType': 'CollectConsumer',
            'argOperatorId': inputid
        }


class MyriaHyperShuffle(algebra.HyperCubeShuffle, MyriaOperator):
    """Represents a HyperShuffle shuffle operator"""
    def compileme(self, inputsym):
        raise NotImplementedError('shouldn''t ever get here, should be turned into HCSP-HCSC pair')  # noqa


class MyriaHyperShuffleProducer(algebra.UnaryOperator, MyriaOperator):
    """A Myria HyperShuffleProducer"""
    def __init__(self, input, hashed_columns,
                 hyper_cube_dims, mapped_hc_dims, cell_partition):
        algebra.UnaryOperator.__init__(self, input)
        self.hashed_columns = hashed_columns
        self.mapped_hc_dimensions = mapped_hc_dims
        self.hyper_cube_dimensions = hyper_cube_dims
        self.cell_partition = cell_partition

    def num_tuples(self):
        return self.input.num_tuples()

    def shortStr(self):
        mapping = {i: '*' for i in range(len(self.hyper_cube_dimensions))}
        mapping.update({h: 'h({col})'.format(col=i)
                        for i, h in zip(self.hashed_columns,
                                        self.mapped_hc_dimensions)})
        hash_string = ','.join(s for m, s in sorted(mapping.items()))
        return "%s(%s)" % (self.opname(), hash_string)

    def compileme(self, inputsym):
        return {
            "opType": "HyperShuffleProducer",
            "hashedColumns": list(self.hashed_columns),
            "mappedHCDimensions": list(self.mapped_hc_dimensions),
            "hyperCubeDimensions": list(self.hyper_cube_dimensions),
            "cellPartition": self.cell_partition,
            "argChild": inputsym
        }


class MyriaHyperShuffleConsumer(algebra.UnaryOperator, MyriaOperator):
    """A Myria HyperShuffleConsumer"""
    def __init__(self, input):
        algebra.UnaryOperator.__init__(self, input)

    def num_tuples(self):
        return self.input.num_tuples()

    def shortStr(self):
        return "%s" % self.opname()

    def compileme(self, inputsym):
        return {
            "opType": "HyperShuffleConsumer",
            "argOperatorId": inputsym
        }


class MyriaQueryScan(algebra.ZeroaryOperator, MyriaOperator):
    """A Myria Query Scan"""
    def __init__(self, sql, scheme, num_tuples=algebra.DEFAULT_CARDINALITY):
        self.sql = str(sql)
        self._scheme = scheme
        self._num_tuples = num_tuples

    def __repr__(self):
        return ("{op}({sql!r}, {sch!r}, {nt!r})"
                .format(op=self.opname(), sql=self.sql,
                        sch=self._scheme, nt=self._num_tuples))

    def num_tuples(self):
        return self._num_tuples

    def shortStr(self):
        return "MyriaQueryScan({sql!r})".format(sql=self.sql)

    def scheme(self):
        return self._scheme

    def compileme(self):
        return {
            "opType": "DbQueryScan",
            "sql": self.sql,
            "schema": scheme_to_schema(self._scheme)
        }


class BreakShuffle(rules.Rule):
    def fire(self, expr):
        if not isinstance(expr, MyriaShuffle):
            return expr

        producer = MyriaShuffleProducer(expr.input, expr.columnlist)
        consumer = MyriaShuffleConsumer(producer)
        return consumer


class BreakHyperCubeShuffle(rules.Rule):
    def fire(self, expr):
        """
        self.hashed_columns = hashed_columns
        self.mapped_hc_dimensions = mapped_hc_dims
        self.hyper_cube_dimensions = hyper_cube_dims
        self.cell_partition = cell_partition
        """
        if not isinstance(expr, MyriaHyperShuffle):
            return expr
        producer = MyriaHyperShuffleProducer(
            expr.input, expr.hashed_columns, expr.hyper_cube_dimensions,
            expr.mapped_hc_dimensions, expr.cell_partition)
        consumer = MyriaHyperShuffleConsumer(producer)
        return consumer


class BreakCollect(rules.Rule):
    def fire(self, expr):
        if not isinstance(expr, MyriaCollect):
            return expr

        producer = MyriaCollectProducer(expr.input, None)
        consumer = MyriaCollectConsumer(producer)
        return consumer


class BreakBroadcast(rules.Rule):
    def fire(self, expr):
        if not isinstance(expr, algebra.Broadcast):
            return expr

        producer = MyriaBroadcastProducer(expr.input)
        consumer = MyriaBroadcastConsumer(producer)
        return consumer


def check_shuffle_xor(exp):
    """Enforce that neither or both inputs to a binary op are shuffled.

    Return True if the arguments are shuffled; False if they are not;
    or raise a ValueError on xor failure.

    Note that we assume that inputs are shuffled in a compatible way.
    """
    left_shuffle = isinstance(exp.left, algebra.Shuffle)
    right_shuffle = isinstance(exp.right, algebra.Shuffle)

    if left_shuffle and right_shuffle:
        return True
    if left_shuffle or right_shuffle:
        raise ValueError("Must shuffle on both inputs of %s" % exp)
    return False


class ShuffleBeforeSetop(rules.Rule):
    def fire(self, exp):
        if not isinstance(exp, (algebra.Difference, algebra.Intersection)):
            return exp

        def shuffle_after(op):
            cols = [expression.UnnamedAttributeRef(i)
                    for i in range(len(op.scheme()))]
            return algebra.Shuffle(child=op, columnlist=cols)

        if not check_shuffle_xor(exp):
            exp.left = shuffle_after(exp.left)
            exp.right = shuffle_after(exp.right)
        return exp


class ShuffleBeforeJoin(rules.Rule):
    def fire(self, expr):
        # If not a join, who cares?
        if not isinstance(expr, algebra.Join):
            return expr

        # If both have shuffles already, who cares?
        if check_shuffle_xor(expr):
            return expr

        # Figure out which columns go in the shuffle
        left_cols, right_cols = \
            convertcondition(expr.condition,
                             len(expr.left.scheme()),
                             expr.left.scheme() + expr.right.scheme())

        # Left shuffle
        left_cols = [expression.UnnamedAttributeRef(i)
                     for i in left_cols]
        left_shuffle = algebra.Shuffle(expr.left, left_cols)
        # Right shuffle
        right_cols = [expression.UnnamedAttributeRef(i)
                      for i in right_cols]
        right_shuffle = algebra.Shuffle(expr.right, right_cols)

        # Construct the object!
        assert isinstance(expr, algebra.ProjectingJoin)
        if isinstance(expr, algebra.ProjectingJoin):
            return algebra.ProjectingJoin(expr.condition,
                                          left_shuffle, right_shuffle,
                                          expr.output_columns)


class HCShuffleBeforeNaryJoin(rules.Rule):
    def __init__(self, catalog):
        assert isinstance(catalog, Catalog)
        self.catalog = catalog

    @staticmethod
    def reversed_index(child_schemes, conditions):
        """Return the reversed index of join conditions. The reverse index
           specify for each column on each relation, which hypercube dimension
           it is mapped to, -1 means this columns is not in the hyper cube
           (not joined).

        Keyword arguments:
        child_schemes -- schemes of children.
        conditions -- join conditions.
        """
        # make it -1 first
        r_index = [[-1] * len(scheme) for scheme in child_schemes]
        for i, jf_list in enumerate(conditions):
            for jf in jf_list:
                r_index[jf[0]][jf[1]] = i
        return r_index

    @staticmethod
    def workload(dim_sizes, child_sizes, r_index):
        """Compute the workload given a hyper cube size assignment"""
        load = 0.0
        for i, size in enumerate(child_sizes):
            # compute subcube sizes
            scale = 1
            for index in r_index[i]:
                if index != -1:
                    scale = scale * dim_sizes[index]
            # add load per server by child i
            load += float(child_sizes[i]) / float(scale)
        return load

    @staticmethod
    def get_hyper_cube_dim_size(num_server, child_sizes,
                                conditions, r_index):
        """Find the optimal hyper cube dimension sizes using BFS.

        Keyword arguments:
        num_server -- number of servers, this sets upper bound of HC cells.
        child_sizes -- cardinality of each child.
        child_schemes -- schemes of children.
        conditions -- join conditions.
        r_index -- reversed index of join conditions.
        """
        # Helper function: compute the product.
        def product(array):
            return reduce(mul, array, 1)
        # Use BFS to find the best possible assignment.
        this = HCShuffleBeforeNaryJoin
        visited = set()
        toVisit = deque()
        toVisit.append(tuple([1 for _ in conditions]))
        min_work_load = None
        while len(toVisit) > 0:
            dim_sizes = toVisit.pop()
            if ((this.workload(dim_sizes, child_sizes, r_index) <
                    min_work_load) or (min_work_load is None)):
                min_work_load = this.workload(
                    dim_sizes, child_sizes, r_index)
                opt_dim_sizes = dim_sizes
            visited.add(dim_sizes)
            for i, d in enumerate(dim_sizes):
                new_dim_sizes = (dim_sizes[0:i] +
                                 tuple([dim_sizes[i] + 1]) +
                                 dim_sizes[i + 1:])
                if (product(new_dim_sizes) <= num_server
                        and new_dim_sizes not in visited):
                    toVisit.append(new_dim_sizes)
        return opt_dim_sizes, min_work_load

    @staticmethod
    def coord_to_worker_id(coordinate, dim_sizes):
        """Convert coordinate of cell to worker id

        Keyword arguments:
        coordinate -- coordinate of hyper cube cell.
        dim_sizes -- sizes of dimensons of hyper cube.
        """
        assert len(coordinate) == len(dim_sizes)
        ret = 0
        for k, v in enumerate(coordinate):
            ret += v * reduce(mul, dim_sizes[k + 1:], 1)
        return ret

    @staticmethod
    def get_cell_partition(dim_sizes, conditions,
                           child_schemes, child_idx, hashed_columns):
        """Generate the cell_partition for a specific child.

        Keyword arguments:
        dim_sizes -- size of each dimension of the hypercube.
        conditions -- each element is an array of (child_idx, column).
        child_schemes -- schemes of children.
        child_idx -- index of this child.
        hashed_columns -- hashed columns of this child.
        """
        assert len(dim_sizes) == len(conditions)
        # make life a little bit easier
        this = HCShuffleBeforeNaryJoin
        # get reverse index
        r_index = this.reversed_index(child_schemes, conditions)
        # find which dims in hyper cube this relation is involved
        hashed_dims = [r_index[child_idx][col] for col in hashed_columns]
        assert -1 not in hashed_dims
        # group by cell according to their projection on subcube voxel
        cell_partition = defaultdict(list)
        coor_ranges = [list(range(d)) for d in dim_sizes]
        for coordinate in itertools.product(*coor_ranges):
            # project a hypercube cell to a subcube voxel
            voxel = [coordinate[dim] for dim in hashed_dims]
            cell_partition[tuple(voxel)].append(
                this.coord_to_worker_id(coordinate, dim_sizes))
        return [wid for vox, wid in sorted(cell_partition.items())]

    def fire(self, expr):
        def add_hyper_shuffle():
            """ Helper function: put a HyperCube shuffle before each child."""
            # make calling static method easier
            this = HCShuffleBeforeNaryJoin
            # get child schemes
            child_schemes = [op.scheme() for op in expr.children()]
            # convert join conditions from expressions to 2d array
            conditions = convert_nary_conditions(
                expr.conditions, child_schemes)
            # get number of servers from catalog
            num_server = self.catalog.get_num_servers()
            # get estimated cardinalities of children
            child_sizes = [child.num_tuples() for child in expr.children()]
            # get reversed index of join conditions
            r_index = this.reversed_index(child_schemes, conditions)
            # compute optimal dimension sizes
            (dim_sizes, workload) = this.get_hyper_cube_dim_size(
                num_server, child_sizes, conditions, r_index)
            # specify HyperCube shuffle to each child
            new_children = []
            for child_idx, child in enumerate(expr.children()):
                # (mapped hc dimension, column index)
                hashed_fields = [(hc_dim, i)
                                 for i, hc_dim
                                 in enumerate(r_index[child_idx])
                                 if hc_dim != -1]
                mapped_dims, hashed_columns = zip(*sorted(hashed_fields))
                # get cell partition for child i
                cell_partition = this.get_cell_partition(
                    dim_sizes, conditions, child_schemes,
                    child_idx, hashed_columns)
                # generate new children
                new_children.append(
                    algebra.HyperCubeShuffle(
                        child, hashed_columns, mapped_dims,
                        dim_sizes, cell_partition))
            # replace the children
            expr.args = new_children

        # only apply to NaryJoin
        if not isinstance(expr, algebra.NaryJoin):
            return expr
        # check if HC shuffle has been placed before
        shuffled_child = [isinstance(op, algebra.HyperCubeShuffle)
                          for op in list(expr.children())]
        if all(shuffled_child):    # already shuffled
            assert len(expr.children()) > 0
            return expr
        elif any(shuffled_child):
            raise NotImplementedError("NaryJoin is partially shuffled?")
        else:                      # add shuffle and order by
            add_hyper_shuffle()
            return expr


class OrderByBeforeNaryJoin(rules.Rule):
    def fire(self, expr):
        # if not NaryJoin, who cares?
        if not isinstance(expr, algebra.NaryJoin):
            return expr
        ordered_child = sum(
            [1 for child in expr.children()
             if isinstance(child, algebra.OrderBy)])

        # already applied
        if ordered_child == len(expr.children()):
            return expr
        elif ordered_child > 0:
            raise Exception("children are partially ordered? ")

        new_children = []
        for child in expr.children():
            # check: this rule must be applied after shuffle
            assert isinstance(child, algebra.HyperCubeShuffle)
            ascending = [True] * len(child.hashed_columns)
            new_children.append(
                algebra.OrderBy(
                    child, child.hashed_columns, ascending))
        expr.args = new_children
        return expr


class BroadcastBeforeCross(rules.Rule):
    def fire(self, expr):
        # If not a CrossProduct, who cares?
        if not isinstance(expr, algebra.CrossProduct):
            return expr

        if (isinstance(expr.left, algebra.Broadcast) or
                isinstance(expr.right, algebra.Broadcast)):
            return expr

        try:
            # By default, broadcast the smaller child
            if expr.left.num_tuples() < expr.right.num_tuples():
                expr.left = algebra.Broadcast(expr.left)
            else:
                expr.right = algebra.Broadcast(expr.right)
        except NotImplementedError, e:
            # If cardinalities unknown, broadcast the right child
            expr.right = algebra.Broadcast(expr.right)

        return expr


class DecomposeGroupBy(rules.Rule):
    """Convert a logical group by into a two-phase group by.

    The local half of the aggregate before the shuffle step, whereas the remote
    half runs after the shuffle step.

    TODO: omit this optimization if the data is already shuffled, or
    if the cardinality of the grouping keys is high.
    """

    @staticmethod
    def do_transfer(op):
        """Introduce a network transfer before a groupby operation."""

        # Get an array of position references to columns in the child scheme
        child_scheme = op.input.scheme()
        group_fields = [expression.toUnnamed(ref, child_scheme)
                        for ref in op.grouping_list]
        if len(group_fields) == 0:
            # Need to Collect all tuples at once place
            op.input = algebra.Collect(op.input)
        else:
            # Need to Shuffle
            op.input = algebra.Shuffle(op.input, group_fields)

    def fire(self, op):
        # Punt if it's not a group by or we've already converted this into an
        # an instance of MyriaGroupBy
        if op.__class__ != algebra.GroupBy:
            return op

        # Bail early if we have any non-decomposable aggregates
        if not all(x.is_decomposable() for x in op.aggregate_list):
            out_op = MyriaGroupBy()
            out_op.copy(op)
            DecomposeGroupBy.do_transfer(out_op)
            return out_op

        num_grouping_terms = len(op.grouping_list)

        local_emitters = []
        local_statemods = []
        remote_emitters = []
        remote_statemods = []
        finalizer_exprs = []

        # The starting positions for the current local, remote aggregate
        local_output_pos = num_grouping_terms
        remote_output_pos = num_grouping_terms
        requires_finalizer = False

        for agg in op.aggregate_list:
            # Multiple emit arguments can be associated with a single
            # decomposition rule; coalesce them all together.
            state = agg.get_decomposable_state()
            assert state

            ################################
            # Extract the set of emitters and statemods required for the
            # local aggregate.
            ################################

            laggs = state.get_local_emitters()
            local_emitters.extend(laggs)
            local_statemods.extend(state.get_local_statemods())

            ################################
            # Extract the set of emitters and statemods required for the
            # remote aggregate.  Remote expressions must be rebased to
            # remove instances of LocalAggregateOutput
            ################################

            raggs = state.get_remote_emitters()
            raggs = [rebase_local_aggregate_output(x, local_output_pos)
                     for x in raggs]
            remote_emitters.extend(raggs)

            rsms = state.get_remote_statemods()
            for sm in rsms:
                update_expr = rebase_local_aggregate_output(
                    sm.update_expr, local_output_pos)
                remote_statemods.append(
                    StateVar(sm.name, sm.init_expr, update_expr))

            ################################
            # Extract any required finalizers.  These must be rebased to remove
            # instances of RemoteAggregateOutput
            ################################

            finalizer = state.get_finalizer()
            if finalizer is not None:
                requires_finalizer = True
                finalizer_exprs.append(
                    rebase_finalizer(finalizer, remote_output_pos))
            else:
                for i in range(len(raggs)):
                    finalizer_exprs.append(
                        UnnamedAttributeRef(remote_output_pos + i))

            local_output_pos += len(laggs)
            remote_output_pos += len(raggs)

        ################################
        # Glue together the local and remote aggregates:
        # Local => Shuffle => Remote => (optional) Finalizer.
        ################################

        local_gb = MyriaGroupBy(op.grouping_list, local_emitters, op.input,
                                local_statemods)

        grouping_fields = [UnnamedAttributeRef(i)
                           for i in range(num_grouping_terms)]

        remote_gb = MyriaGroupBy(grouping_fields, remote_emitters, local_gb,
                                 remote_statemods)

        DecomposeGroupBy.do_transfer(remote_gb)

        if requires_finalizer:
            # Pass through grouping terms
            gmappings = [(None, UnnamedAttributeRef(i))
                         for i in range(num_grouping_terms)]
            fmappings = [(None, fx) for fx in finalizer_exprs]
            return algebra.Apply(gmappings + fmappings, remote_gb)
        return remote_gb


class AddAppendTemp(rules.Rule):
    def fire(self, op):
        if type(op) is not MyriaStoreTemp:
            return op

        child = op.input
        if type(child) is not MyriaUnionAll:
            return op

        left = child.left
        right = child.right
        rel_name = op.name

        is_scan = lambda op: type(op) is MyriaScanTemp and op.name == rel_name
        if is_scan(left) and not any(is_scan(op) for op in right.walk()):
                return MyriaAppendTemp(name=rel_name, input=right)

        elif is_scan(right) and not any(is_scan(op) for op in left.walk()):
                return MyriaAppendTemp(name=rel_name, input=left)

        return op


class PushIntoSQL(rules.Rule):
    def __init__(self, dialect=None):
        self.dialect = dialect or postgresql.dialect()

    def fire(self, expr):
        if isinstance(expr, (algebra.Scan, algebra.ScanTemp)):
            return expr
        cat = SQLCatalog()
        try:
            sql_plan = cat.get_sql(expr)
            sql_string = sql_plan.compile(dialect=self.dialect)
            sql_string.visit_bindparam = sql_string.render_literal_bindparam
            return MyriaQueryScan(sql=sql_string.process(sql_plan),
                                  scheme=expr.scheme(),
                                  num_tuples=expr.num_tuples())
        except NotImplementedError, e:
            LOGGER.warn("Error converting {plan}: {e}"
                        .format(plan=expr, e=e))
            return expr


class MergeToNaryJoin(rules.Rule):
    """Merge consecutive binary join into a single multiway join
    Note: this code assumes that the binary joins form a left deep tree
    before the merge."""
    @staticmethod
    def mergable(op):
        """Recursively checks whether an operator is mergable to NaryJoin.
        An operator will be merged to NaryJoin if its subtree contains
        only joins.
        """
        allowed_intermediate_types = (algebra.ProjectingJoin, algebra.Select)
        if issubclass(type(op), algebra.ZeroaryOperator):
            return True
        if not isinstance(op, allowed_intermediate_types):
            return False
        elif issubclass(type(op), algebra.UnaryOperator):
            return MergeToNaryJoin.mergable(op.input)
        elif issubclass(type(op), algebra.BinaryOperator):
            return (MergeToNaryJoin.mergable(op.left) and
                    MergeToNaryJoin.mergable(op.right))

    @staticmethod
    def collect_join_groups(op, conditions, children):
        assert isinstance(op, algebra.ProjectingJoin)
        assert (isinstance(op.right, algebra.Select)
                or issubclass(type(op.right), algebra.ZeroaryOperator))
        children.append(op.right)
        conjuncs = expression.extract_conjuncs(op.condition)
        for cond in conjuncs:
            conditions.get_or_insert(cond.left)
            conditions.get_or_insert(cond.right)
            conditions.union(cond.left, cond.right)
        scan_then_select = (isinstance(op.left, algebra.Select) and
                            isinstance(op.left.input, algebra.ZeroaryOperator))
        if (scan_then_select or
                issubclass(type(op.left), algebra.ZeroaryOperator)):
            children.append(op.left)
        else:
            assert isinstance(op.left, algebra.ProjectingJoin)
            MergeToNaryJoin.collect_join_groups(op.left, conditions, children)

    def fire(self, op):
        if not isinstance(op, algebra.ProjectingJoin):
            return op

        # if op is the only binary join, return
        if not isinstance(op.left, algebra.ProjectingJoin):
            return op
        # if it is not mergable, e.g. aggregation along the path, return
        if not MergeToNaryJoin.mergable(op):
            return op
        # do the actual merge
        # 1. collect join groups
        join_groups = UnionFind()
        children = []
        MergeToNaryJoin.collect_join_groups(
            op, join_groups, children)
        # 2. extract join groups from the union-find data structure
        join_conds = defaultdict(list)
        for field, key in join_groups.parents.items():
            join_conds[key].append(field)
        conditions = [v for (k, v) in join_conds.items()]
        # Note: a cost based join order optimization need to be implemented.
        ordered_conds = sorted(conditions, key=lambda cond: min(cond))
        # 3. reverse the children due to top-down tree traversal
        return algebra.NaryJoin(
            list(reversed(children)), ordered_conds, op.output_columns)


class GetCardinalities(rules.Rule):
    """ get cardinalities information of Zeroary operators.
    """
    def __init__(self, catalog):
        assert isinstance(catalog, Catalog)
        self.catalog = catalog

    def fire(self, expr):
        # if not Zeroary operator, who cares?
        if not issubclass(type(expr), algebra.ZeroaryOperator):
            return expr

        if issubclass(type(expr), algebra.Scan):
            rel = expr.relation_key
            expr._cardinality = self.catalog.num_tuples(rel)
            return expr
        expr._cardinality = algebra.DEFAULT_CARDINALITY
        return expr


# 6. shuffle logics, hyper_cube_shuffle_logic is only used in HCAlgebra
left_deep_tree_shuffle_logic = [
    ShuffleBeforeSetop(),
    ShuffleBeforeJoin(),
    BroadcastBeforeCross()
]

# 7. distributed groupby
# this need to be put after shuffle logic
distributed_group_by = [
    # DecomposeGroupBy may introduce a complex GroupBy,
    # so we must run SimpleGroupBy after it. TODO no one likes this.
    DecomposeGroupBy(),
    rules.SimpleGroupBy(),
    rules.CountToCountall(),   # TODO revisit when we have NULL support.
    rules.DedupGroupBy(),
    rules.EmptyGroupByToDistinct(),
]

# 8. Myriafy logical operators
# replace logical operator with its corresponding Myria operators
myriafy = [
    rules.OneToOne(algebra.CrossProduct, MyriaCrossProduct),
    rules.OneToOne(algebra.Store, MyriaStore),
    rules.OneToOne(algebra.StoreTemp, MyriaStoreTemp),
    rules.OneToOne(algebra.StatefulApply, MyriaStatefulApply),
    rules.OneToOne(algebra.Apply, MyriaApply),
    rules.OneToOne(algebra.Select, MyriaSelect),
    rules.OneToOne(algebra.Distinct, MyriaDupElim),
    rules.OneToOne(algebra.Shuffle, MyriaShuffle),
    rules.OneToOne(algebra.HyperCubeShuffle, MyriaHyperShuffle),
    rules.OneToOne(algebra.Collect, MyriaCollect),
    rules.OneToOne(algebra.ProjectingJoin, MyriaSymmetricHashJoin),
    rules.OneToOne(algebra.NaryJoin, MyriaLeapFrogJoin),
    rules.OneToOne(algebra.Scan, MyriaScan),
    rules.OneToOne(algebra.ScanTemp, MyriaScanTemp),
    rules.OneToOne(algebra.SingletonRelation, MyriaSingleton),
    rules.OneToOne(algebra.EmptyRelation, MyriaEmptyRelation),
    rules.OneToOne(algebra.UnionAll, MyriaUnionAll),
    rules.OneToOne(algebra.Difference, MyriaDifference),
    rules.OneToOne(algebra.OrderBy, MyriaInMemoryOrderBy),
]

# 9. break communication boundary
# get producer/consumer pair
break_communication = [
    BreakHyperCubeShuffle(),
    BreakShuffle(),
    BreakCollect(),
    BreakBroadcast(),
]


class MyriaAlgebra(Algebra):
    """ Myria algebra abstract class"""
    language = MyriaLanguage

    fragment_leaves = (
        MyriaShuffleConsumer,
        MyriaCollectConsumer,
        MyriaBroadcastConsumer,
        MyriaHyperShuffleConsumer,
        MyriaScan,
        MyriaScanTemp
    )


class MyriaLeftDeepTreeAlgebra(MyriaAlgebra):
    """Myria physical algebra using left deep tree pipeline and 1-D shuffle"""
    def opt_rules(self, **kwargs):
        opt_grps_sequence = [
            rules.remove_trivial_sequences,
            [
                rules.SimpleGroupBy(),
                rules.CountToCountall(),  # TODO revisit when we have NULLs
                rules.ProjectToDistinctColumnSelect(),
                rules.DistinctToGroupBy(),
                rules.DedupGroupBy(),
            ],
            rules.push_select,
            rules.push_project,
            rules.push_apply,
            left_deep_tree_shuffle_logic,
            distributed_group_by,
            [rules.PushApply()],
        ]

        if kwargs.get('push_sql', False):
            opt_grps_sequence.append([
                PushIntoSQL(dialect=kwargs.get('dialect'))])

        compile_grps_sequence = [
            myriafy,
            [AddAppendTemp()],
            break_communication
        ]

        rule_grps_sequence = opt_grps_sequence + compile_grps_sequence
        return list(itertools.chain(*rule_grps_sequence))


class MyriaHyperCubeAlgebra(MyriaAlgebra):
    """Myria physical algebra using HyperCubeShuffle and LeapFrogJoin"""
    def opt_rules(self, **kwargs):
        # this rule is hyper cube shuffle specific
        merge_to_nary_join = [
            MergeToNaryJoin()
        ]

        # catalog aware hc shuffle rules, so put them here
        hyper_cube_shuffle_logic = [
            GetCardinalities(self.catalog),
            HCShuffleBeforeNaryJoin(self.catalog),
            OrderByBeforeNaryJoin(),
        ]

        opt_grps_sequence = [
            rules.remove_trivial_sequences,
            [
                rules.SimpleGroupBy(),
                # TODO revisit when we have NULL support.
                rules.CountToCountall(),
                rules.DistinctToGroupBy(),
                rules.DedupGroupBy(),
            ],
            rules.push_select,
            rules.push_project,
            merge_to_nary_join,
            rules.push_apply,
            left_deep_tree_shuffle_logic,
            distributed_group_by,
            hyper_cube_shuffle_logic
        ]

        if kwargs.get('push_sql', False):
            opt_grps_sequence.append([PushIntoSQL()])

        compile_grps_sequence = [
            myriafy,
            [AddAppendTemp()],
            break_communication
        ]
        rule_grps_sequence = opt_grps_sequence + compile_grps_sequence
        return list(itertools.chain(*rule_grps_sequence))

    def __init__(self, catalog=None):
        self.catalog = catalog


class OpIdFactory(object):
    def __init__(self):
        self.count = 0

    def alloc(self):
        ret = self.count
        self.count += 1
        return ret

    def getter(self):
        return lambda: self.alloc()


def ensure_store_temp(label, op):
    """Returns a StoreTemp that assigns the given operator to a temp relation
    with the given name. Op must be a 'normal operator', i.e.,
    not a Store/StoreTemp or a control-flow sub-plan operator."""
    assert not isinstance(op, (algebra.Store, algebra.StoreTemp))
    assert not isinstance(op, (algebra.Sequence, algebra.Parallel,
                               algebra.DoWhile))
    assert isinstance(label, basestring) and len(label) > 0
    return MyriaStoreTemp(input=op, name=label)


def compile_fragment(frag_root):
    """Given a root operator, produce a SubQueryEncoding."""

    # A dictionary mapping each object to a unique, object-dependent id.
    # Since we want this to be truly unique for each object instance, even if
    # two objects are equal, we use id(obj) as the key.
    opid_factory = OpIdFactory()
    op_ids = defaultdict(opid_factory.getter())

    def one_fragment(rootOp):
        """Given an operator that is the root of a query fragment/plan, extract
        the operators in the fragment. Assembles a list cur_frag of the
        operators in the current fragment, in preorder from the root.

        This operator also assembles a queue of the discovered roots of later
        fragments, e.g., when there is a ShuffleProducer below. The list of
        operators that should be treated as fragment leaves is given by
        MyriaAlgebra.fragment_leaves. """

        # The current fragment starts with the current root
        cur_frag = [rootOp]
        # Initially, there are no new roots discovered below leaves of this
        # fragment.
        queue = []
        if isinstance(rootOp, MyriaAlgebra.fragment_leaves):
            # The current root operator is a fragment leaf, such as a
            # ShuffleProducer. Append its children to the queue of new roots.
            for child in rootOp.children():
                queue.append(child)
        else:
            # Otherwise, the children belong in this fragment. Recursively go
            # discover their fragments, including the queue of roots below
            # their children.
            for child in rootOp.children():
                (child_frag, child_queue) = one_fragment(child)
                # Add their fragment onto this fragment
                cur_frag += child_frag
                # Add their roots-of-next-fragments into our queue
                queue += child_queue
        return (cur_frag, queue)

    def fragments(rootOp):
        """Given the root of a query plan, recursively determine all the
        fragments in it."""
        # The queue of fragment roots. Initially, just the root of this query
        queue = [rootOp]
        ret = []
        while len(queue) > 0:
            # Get the next fragment root
            rootOp = queue.pop(0)
            # .. recursively learn the entire fragment, and any newly
            # discovered roots.
            (op_frag, op_queue) = one_fragment(rootOp)
            # .. Myria JSON expects the fragment operators in reverse order,
            # i.e., root at the bottom.
            ret.append(reversed(op_frag))
            # .. and collect the newly discovered fragment roots.
            queue.extend(op_queue)
        return ret

    def call_compile_me(op):
        "A shortcut to call the operator's compile_me function."
        op_id = op_ids[id(op)]
        child_op_ids = [op_ids[id(child)] for child in op.children()]
        op_dict = op.compileme(*child_op_ids)
        op_dict['opName'] = op.shortStr()
        assert isinstance(op_id, int), (type(op_id), op_id)
        op_dict['opId'] = op_id
        return op_dict

    # Determine and encode the fragments.
    return [{'operators': [call_compile_me(op) for op in frag]}
            for frag in fragments(frag_root)]


def compile_plan(plan_op):
    subplan_ops = (algebra.Parallel, algebra.Sequence, algebra.DoWhile)
    if not isinstance(plan_op, subplan_ops):
        plan_op = algebra.Parallel([plan_op])

    if isinstance(plan_op, algebra.Parallel):
        frag_list = [compile_fragment(op) for op in plan_op.children()]
        return {"type": "SubQuery",
                "fragments": list(itertools.chain(*frag_list))}

    elif isinstance(plan_op, algebra.Sequence):
        plan_list = [compile_plan(pl_op) for pl_op in plan_op.children()]
        return {"type": "Sequence", "plans": plan_list}

    elif isinstance(plan_op, algebra.DoWhile):
        children = plan_op.children()
        if len(children) < 2:
            raise ValueError('DoWhile must have at >= 2 children: body and condition')  # noqa
        condition = children[-1]
        if isinstance(condition, subplan_ops):
            raise ValueError('DoWhile condition cannot be a subplan op {cls}'.format(cls=condition.__class__))  # noqa
        condition = ensure_store_temp('__dowhile_{}_condition'.format(id(
            plan_op)), condition)
        plan_op.args = children[:-1] + [condition]
        body = [compile_plan(pl_op) for pl_op in plan_op.children()]
        return {"type": "DoWhile",
                "body": body,
                "condition": condition.name}

    raise NotImplementedError("compiling subplan op {}".format(type(plan_op)))


def compile_to_json(raw_query, logical_plan, physical_plan,
                    language="not specified"):
    """This function compiles a physical query plan to the JSON suitable for
    submission to the Myria REST API server. The logical plan is converted to a
    string and passed along unchanged."""

    # Store/StoreTemp is a reasonable physical plan... for now.
    if isinstance(physical_plan, (algebra.Store, algebra.StoreTemp)):
        physical_plan = algebra.Parallel([physical_plan])

    subplan_ops = (algebra.Parallel, algebra.Sequence, algebra.DoWhile)
    assert isinstance(physical_plan, subplan_ops), \
        'Physical plan must be a subplan operator, not {}'.format(type(physical_plan))  # noqa

    # raw_query must be a string
    if not isinstance(raw_query, basestring):
        raise ValueError("raw query must be a string")

    return {"rawQuery": raw_query,
            "logicalRa": str(logical_plan),
            "language": language,
            "plan": compile_plan(physical_plan)}



"""Representation of a Myria relation key.

Myria relations are identified by a tuple of user, program, relation_name."""


class RelationKey(object):
    def __init__(self, *args):
        if len(args) == 1:
            self.user = "public"
            self.program = "adhoc"
            self.relation = args[0]
        else:
            self.user, self.program, self.relation = args
        assert self.user and isinstance(self.user, basestring)
        assert self.program and isinstance(self.program, basestring)
        assert self.relation and isinstance(self.relation, basestring)

    def __repr__(self):
        return 'RelationKey(%r,%r,%r)' % (self.user, self.program,
                                          self.relation)

    def __str__(self):
        return '%s:%s:%s' % (self.user, self.program, self.relation)

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def __hash__(self):
        return hash(str(self))

    @classmethod
    def from_string(cls, s):
        """Create a RelationKey from a colon-delimited string."""
        toks = s.split(':')
        assert len(toks) <= 3

        return cls(*toks)



"""
Aggregate expressions for use in Raco
"""

from .expression import *
from .function import UnaryFunction, SQRT, POW
from .statevar import *
from raco import types
from abc import abstractmethod
import math


class DecomposableAggregateState(object):
    """State associated with decomposable aggregates.

    :param local_emitters: A list of one or more aggregates to run prior to
    the shuffle operation
    :param local_statemods: A list of StateVars associated with local aggs
    :param remote_emitters: A list of one or more aggregates to run after
    the shuffle operation
    :param remote_statemods: A list of StateVars associated with remote aggs
    :param finalizer: An optional expression that reduces the remote aggregate
    outputs to a single value.
    """
    def __init__(self, local_emitters=(), local_statemods=(),
                 remote_emitters=(), remote_statemods=(), finalizer=None):

        assert all(isinstance(a, AggregateExpression) for a in local_emitters)
        assert all(isinstance(a, AggregateExpression) for a in remote_emitters)
        assert all(isinstance(s, StateVar) for s in local_statemods)
        assert all(isinstance(s, StateVar) for s in remote_statemods)
        if finalizer is not None:
            assert isinstance(finalizer, Expression)
            assert not isinstance(finalizer, AggregateExpression)

        self.local_emitters = local_emitters
        self.local_statemods = local_statemods
        self.remote_emitters = remote_emitters
        self.remote_statemods = remote_statemods
        self.finalizer = finalizer

    def get_local_emitters(self):
        return self.local_emitters

    def get_remote_emitters(self):
        return self.remote_emitters

    def get_local_statemods(self):
        return self.local_statemods

    def get_remote_statemods(self):
        return self.remote_statemods

    def get_finalizer(self):
        return self.finalizer


class AggregateExpression(Expression):
    def is_decomposable(self):
        return self.get_decomposable_state() is not None

    @abstractmethod
    def get_decomposable_state(self):
        pass


class BuiltinAggregateExpression(AggregateExpression):
    def evaluate(self, _tuple, scheme, state=None):
        raise NotImplementedError("{expr}.evaluate".format(expr=type(self)))

    def get_decomposable_state(self):
        return None

    @abstractmethod
    def evaluate_aggregate(self, tuple_iterator, scheme):
        """Evaluate an aggregate over a bag of tuples"""


class UdaAggregateExpression(AggregateExpression, UnaryOperator):
    """A user-defined aggregate.

    A UDA wraps an emit expression that is responsible for emitting a
    value for each tuple group.
    """
    def __init__(self, emitter, decomposable_state=None):
        UnaryOperator.__init__(self, emitter)
        self.decomposable_state = decomposable_state

    def evaluate(self, _tuple, scheme, state=None):
        """Evaluate the UDA sub-expression.

        Note that the emitter should only reference the state argument.
        """
        return self.input.evaluate(None, None, state)

    def typeof(self, scheme, state_scheme):
        return self.input.typeof(None, state_scheme)

    def get_decomposable_state(self):
        return self.decomposable_state

    def set_decomposable_state(self, ds):
        self.decomposable_state = ds

    def __repr__(self):
        return "{op}({se!r})".format(op=self.opname(), se=self.input)

    def __str__(self):
        return 'UDA(%s)' % self.input


class LocalAggregateOutput(ZeroaryOperator):
    """Dummy placeholder to refer to the output of a local aggregate."""
    def __init__(self, index=0):
        """Initialize a LocalAggregateOutput

        :param index: The index into the array of local aggregate outputs.
        """
        self.index = index

    def evaluate(self, _tuple, scheme, state=None):
        """Raise an error on attempted evaluation.

        These expressions should be compiled away prior to evaluation.
        """
        raise NotImplementedError()

    def typeof(self, scheme, state_scheme):
        raise NotImplementedError()  # See above comment


class RemoteAggregateOutput(ZeroaryOperator):
    """Dummy placeholder to refer to the output of a merge aggregate."""
    def __init__(self, index):
        """Instantiate a merge aggregate object.

        index is the position relative to the start of the remote aggregate.
        """
        self.index = index

    def evaluate(self, _tuple, scheme, state=None):
        """Raise an error on attempted evaluation.

        These expressions should be compiled away prior to evaluation.
        """
        raise NotImplementedError()

    def typeof(self, scheme, state_scheme):
        raise NotImplementedError()  # See above comment


def rebase_local_aggregate_output(expr, offset):
    """Convert LocalAggregateOutput instances to raw column references."""
    assert isinstance(expr, Expression)

    def convert(n):
        if isinstance(n, LocalAggregateOutput):
            return UnnamedAttributeRef(n.index + offset)
        n.apply(convert)
        return n
    return convert(expr)


def rebase_finalizer(expr, offset):
    """Convert RemoteAggregateOutput instances to raw column references."""
    assert isinstance(expr, Expression)

    def convert(n):
        if isinstance(n, RemoteAggregateOutput):
            return UnnamedAttributeRef(n.index + offset)
        n.apply(convert)
        return n
    return convert(expr)


class TrivialAggregateExpression(BuiltinAggregateExpression):
    """A built-in aggregate with identical local and remote aggregates."""

    def get_decomposable_state(self):
        return DecomposableAggregateState(
            [self], [], [self.__class__(LocalAggregateOutput(0))], [])

    def is_decomposable(self):
        return True


class MAX(UnaryFunction, TrivialAggregateExpression):

    def evaluate_aggregate(self, tuple_iterator, scheme):
        inputs = (self.input.evaluate(t, scheme) for t in tuple_iterator)
        return max(inputs)

    def typeof(self, scheme, state_scheme):
        return self.input.typeof(scheme, state_scheme)


class MIN(UnaryFunction, TrivialAggregateExpression):
    def evaluate_aggregate(self, tuple_iterator, scheme):
        inputs = (self.input.evaluate(t, scheme) for t in tuple_iterator)
        return min(inputs)

    def typeof(self, scheme, state_scheme):
        return self.input.typeof(scheme, state_scheme)


class COUNTALL(ZeroaryOperator, BuiltinAggregateExpression):
    def evaluate_aggregate(self, tuple_iterator, scheme):
        return len(tuple_iterator)

    def typeof(self, scheme, state_scheme):
        return types.LONG_TYPE

    def get_decomposable_state(self):
        return DecomposableAggregateState(
            [self], [], [SUM(LocalAggregateOutput(0))], [])

    def is_decomposable(self):
        return True


class COUNT(UnaryFunction, BuiltinAggregateExpression):
    def evaluate_aggregate(self, tuple_iterator, scheme):
        inputs = (self.input.evaluate(t, scheme) for t in tuple_iterator)
        count = 0
        for t in inputs:
            if t is not None:
                count += 1
        return count

    def typeof(self, scheme, state_scheme):
        return types.LONG_TYPE

    def get_decomposable_state(self):
        return DecomposableAggregateState(
            [self], [], [SUM(LocalAggregateOutput(0))], [])

    def is_decomposable(self):
        return True


class SUM(UnaryFunction, TrivialAggregateExpression):
    def evaluate_aggregate(self, tuple_iterator, scheme):
        inputs = (self.input.evaluate(t, scheme) for t in tuple_iterator)

        return sum(x for x in inputs if x is not None)

    def typeof(self, scheme, state_scheme):
        input_type = self.input.typeof(scheme, state_scheme)
        check_is_numeric(input_type)
        return input_type


class AVG(UnaryFunction, BuiltinAggregateExpression):
    def evaluate_aggregate(self, tuple_iterator, scheme):
        inputs = (self.input.evaluate(t, scheme) for t in tuple_iterator)
        filtered = list(x for x in inputs if x is not None)
        return sum(filtered) / len(filtered)

    def typeof(self, scheme, state_scheme):
        input_type = self.input.typeof(scheme, state_scheme)
        check_is_numeric(input_type)
        return types.DOUBLE_TYPE

    def get_decomposable_state(self):
        return DecomposableAggregateState(
            [SUM(self.input), COUNT(self.input)], [],
            [SUM(LocalAggregateOutput(0)), SUM(LocalAggregateOutput(1))], [],
            DIVIDE(RemoteAggregateOutput(0), RemoteAggregateOutput(1)))

    def is_decomposable(self):
        return True


class STDEV(UnaryFunction, BuiltinAggregateExpression):
    def evaluate_aggregate(self, tuple_iterator, scheme):
        inputs = (self.input.evaluate(t, scheme) for t in tuple_iterator)
        filtered = [x for x in inputs if x is not None]

        n = len(filtered)
        if n < 2:
            return 0.0

        mean = float(sum(filtered)) / n
        return math.sqrt(sum((a - mean) ** 2 for a in filtered) / n)

    def typeof(self, scheme, state_scheme):
        input_type = self.input.typeof(scheme, state_scheme)
        check_is_numeric(input_type)
        return types.DOUBLE_TYPE

    def get_decomposable_state(self):
        si = self.input
        remits = [SUM(LocalAggregateOutput(i)) for i in range(3)]

        _sum = RemoteAggregateOutput(0)
        ssq = RemoteAggregateOutput(1)
        count = RemoteAggregateOutput(2)
        finalizer = SQRT(MINUS(DIVIDE(ssq, count),
                         POW(DIVIDE(_sum, count),
                             NumericLiteral(2))))

        return DecomposableAggregateState(
            [SUM(si), SUM(TIMES(si, si)), COUNT(si)], [],
            remits, [], finalizer)

    def is_decomposable(self):
        return True



from raco import algebra


def graph_to_dot(graph, **kwargs):
    """Graph is expected to be a dict of the form { 'nodes' : list(), 'edges' :
    list() }. This function returns a string that will be input to dot."""

    title = kwargs.get('title', '')

    # Template, including setup and formatting:
    template = """digraph G {
      ratio = "fill" ;
      size = "4.0, 4.0" ;
      page = "4, 4" ;
      margin = 0 ;
      mincross = 2.0 ;
      rankdir = "BT" ;
      nodesep = 0.25 ;
      ranksep = 0.25 ;
      node [fontname="Helvetica", fontsize=10,
            shape=oval, style=filled, fillcolor=white ] ;

      // The nodes
      %s

      // The edges
      %s

      // The title
      labelloc="t";
      label="%s";
}"""

    # Nodes:
    nodes = ['"%s" [label="%s"] ;' % (id(n), n.shortStr().replace(r'"', r'\"'))
             for n in graph['nodes']]
    node_str = '\n      '.join(nodes)

    # Edges:
    edges = ['"%s" -> "%s" ;' % (id(x), id(y)) for (x, y) in graph['edges']]
    edge_str = '\n      '.join(edges)

    return template % (node_str, edge_str, title)


def operator_to_dot(operator, graph=None, **kwargs):
    """Operator is expected to be an object of class raco.algebra.Operator"""
    graph = operator.collectGraph(graph)
    return graph_to_dot(graph, **kwargs)


def get_dot(obj):
    if isinstance(obj, dict) and 'nodes' in dict and 'edges' in dict:
        return graph_to_dot(obj)
    elif isinstance(obj, algebra.Operator):
        return operator_to_dot(obj)
    raise NotImplementedError('Unable to get dot from object type %s' % type(obj))  # noqa




import collections

import raco.scheme as scheme
import raco.myrial.myrial_test as myrial_test
from raco.myrial.exceptions import *
from raco import types


class SetopTestFunctions(myrial_test.MyrialTestCase):

    emp_table1 = collections.Counter([
        (1, 2, "Bill Howe", 25000),
        (1, 2, "Bill Howe", 25000),
        (2, 1, "Dan Halperin", 90000),
        (3, 1, "Andrew Whitaker", 5000),
        (3, 1, "Andrew Whitaker", 5000),
        (4, 2, "Shumo Chu", 5000),
        (5, 1, "Victor Almeida", 25000),
        (6, 3, "Dan Suciu", 90000),
        (7, 1, "Magdalena Balazinska", 25000)])

    emp_key1 = "andrew:adhoc:employee1"

    emp_table2 = collections.Counter([
        (1, 2, "Bill Howe", 25000),
        (7, 1, "Magdalena Balazinska", 25000),
        (7, 1, "Magdalena Balazinska", 25000),
        (8, 2, "JingJing Wang", 47000)])

    emp_key2 = "andrew:adhoc:employee2"

    emp_schema = scheme.Scheme([("id", types.LONG_TYPE),
                                ("dept_id", types.LONG_TYPE),
                                ("name", types.STRING_TYPE),
                                ("salary", types.LONG_TYPE)])

    def setUp(self):
        super(SetopTestFunctions, self).setUp()

        self.db.ingest(SetopTestFunctions.emp_key1,
                       SetopTestFunctions.emp_table1,
                       SetopTestFunctions.emp_schema)

        self.db.ingest(SetopTestFunctions.emp_key2,
                       SetopTestFunctions.emp_table2,
                       SetopTestFunctions.emp_schema)

    def test_unionall(self):
        query = """
        out = UNIONALL(SCAN(%s), SCAN(%s));
        STORE(out, OUTPUT);
        """ % (self.emp_key1, self.emp_key2)

        expected = self.emp_table1 + self.emp_table2
        self.check_result(query, expected)

    def test_unionall_schema_mismatch(self):
        query = """
        T1 = [FROM SCAN(%s) AS X EMIT id, dept_id, name, salary, 7 as seven];
        out = UNIONALL(T1, SCAN(%s));
        STORE(out, OUTPUT);
        """ % (self.emp_key1, self.emp_key2)

        with self.assertRaises(SchemaMismatchException):
            self.get_logical_plan(query)

    def test_unionall_inline(self):
        query = """
        out = SCAN(%s) + SCAN(%s);
        STORE(out, OUTPUT);
        """ % (self.emp_key1, self.emp_key2)

        expected = self.emp_table1 + self.emp_table2
        self.check_result(query, expected)

    def test_unionall_inline_ternary(self):
        query = """
        out = SCAN(%s) + [FROM SCAN(%s) AS X EMIT *] + SCAN(%s);
        STORE(out, OUTPUT);
        """ % (self.emp_key1, self.emp_key1, self.emp_key1)

        expected = self.emp_table1 + self.emp_table1 + self.emp_table1
        self.check_result(query, expected)

    def test_diff1(self):
        query = """
        out = DIFF(SCAN(%s), SCAN(%s));
        STORE(out, OUTPUT);
        """ % (self.emp_key1, self.emp_key2)

        expected = collections.Counter(
            set(self.emp_table1).difference(set(self.emp_table2)))
        self.check_result(query, expected)

    def test_diff2(self):
        query = """
        out = DIFF(SCAN(%s), SCAN(%s));
        STORE(out, OUTPUT);
        """ % (self.emp_key2, self.emp_key1)

        expected = collections.Counter(
            set(self.emp_table2).difference(set(self.emp_table1)))
        self.check_result(query, expected)

    def test_diff_schema_mismatch(self):
        query = """
        T1 = [FROM SCAN(%s) AS X EMIT id, dept_id, name];
        out = DIFF(SCAN(%s), T1);
        STORE(out, OUTPUT);
        """ % (self.emp_key1, self.emp_key2)

        with self.assertRaises(SchemaMismatchException):
            self.get_logical_plan(query)

    def test_diff_while_schema_mismatch(self):
        query = """
        Orig = [2 as x];
        T1 = [2 as x];
        do
          Bad = diff(T1, Orig);
          T1 = [3 as x, 3 as y];
        while [from Bad emit count(*) > 0];
        store(T1, OUTPUT);
        """

        with self.assertRaises(SchemaMismatchException):
            # TODO Even if executed, this test does not throw exception
            self.get_logical_plan(query)

    def test_diff_while_schema_mismatch2(self):
        query = """
        Orig = [2 as x];
        T1 = [3 as x];
        do
          Bad = diff(T1, Orig);
          T1 = [3 as x, 3 as y];
        while [from Bad emit count(*) > 0];
        store(T1, OUTPUT);
        """

        with self.assertRaises(SchemaMismatchException):
            # TODO If executed, this test loops infinitely
            self.get_logical_plan(query)

    def test_intersect1(self):
        query = """
        out = INTERSECT(SCAN(%s), SCAN(%s));
        STORE(out, OUTPUT);
        """ % (self.emp_key1, self.emp_key2)

        expected = collections.Counter(
            set(self.emp_table2).intersection(set(self.emp_table1)))
        self.check_result(query, expected, skip_json=True)

    def test_intersect2(self):
        query = """
        out = INTERSECT(SCAN(%s), SCAN(%s));
        STORE(out, OUTPUT);
        """ % (self.emp_key2, self.emp_key1)

        expected = collections.Counter(
            set(self.emp_table1).intersection(set(self.emp_table2)))
        self.check_result(query, expected, skip_json=True)

    def test_intersect_schema_mismatch(self):
        query = """
        T1 = [FROM SCAN(%s) AS X EMIT id, dept_id, name];
        out = INTERSECT(T1, SCAN(%s));
        STORE(out, OUTPUT);
        """ % (self.emp_key1, self.emp_key2)

        with self.assertRaises(SchemaMismatchException):
            self.get_logical_plan(query)



from abc import ABCMeta, abstractmethod
import raco.expression as expression

import logging
LOG = logging.getLogger(__name__)


class Algebra(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def opt_rules(self, **kwargs):
        raise NotImplementedError("{op}.opt_rules()".format(op=type(self)))


class Language(object):
    __metaclass__ = ABCMeta

    # By default, reuse scans
    reusescans = True

    @staticmethod
    def preamble(query=None, plan=None):
        return ""

    @staticmethod
    def postamble(query=None, plan=None):
        return ""

    @staticmethod
    def body(compileResult):
        return compileResult

    @classmethod
    def compile_stringliteral(cls, value):
        return '"%s"' % value

    @staticmethod
    def log(txt):
        """Emit code that will generate a log message at runtime. Defaults to
        nothing."""
        return ""

    @classmethod
    def compile_numericliteral(cls, value):
        return '%s' % value

    @classmethod
    def compile_attribute(cls, attr):
        return attr.compile()

    @classmethod
    def conjunction(cls, *args):
        return cls.expression_combine(args, operator="and")

    @classmethod
    def disjunction(cls, *args):
        return cls.expression_combine(args, operator="or")

    @classmethod
    def compile_expression(cls, expr):
        compilevisitor = CompileExpressionVisitor(cls)
        expr.accept(compilevisitor)
        return compilevisitor.getresult()

    @classmethod
    @abstractmethod
    def expression_combine(cls, args, operator="and"):
        """Combine the given arguments using the specified infix operator"""


class CompileExpressionVisitor(expression.ExpressionVisitor):
    def __init__(self, language):
        self.language = language
        self.combine = language.expression_combine
        self.stack = []

    def getresult(self):
        assert len(self.stack) == 1
        return self.stack.pop()

    def __visit_BinaryOperator__(self, binaryexpr):
        right = self.stack.pop()
        left = self.stack.pop()
        return left, right

    def visit_NOT(self, unaryexpr):
        inputexpr = self.stack.pop()
        self.stack.append(self.language.negation(inputexpr))

    def visit_AND(self, binaryexpr):
        left, right = self.__visit_BinaryOperator__(binaryexpr)
        self.stack.append(self.language.conjunction(left, right))

    def visit_OR(self, binaryexpr):
        left, right = self.__visit_BinaryOperator__(binaryexpr)
        self.stack.append(self.language.disjunction(left, right))

    def visit_EQ(self, binaryexpr):
        left, right = self.__visit_BinaryOperator__(binaryexpr)
        self.stack.append(self.combine([left, right], operator="=="))

    def visit_NEQ(self, binaryexpr):
        left, right = self.__visit_BinaryOperator__(binaryexpr)
        self.stack.append(self.combine([left, right], operator="!="))

    def visit_GT(self, binaryexpr):
        left, right = self.__visit_BinaryOperator__(binaryexpr)
        self.stack.append(self.combine([left, right], operator=">"))

    def visit_LT(self, binaryexpr):
        left, right = self.__visit_BinaryOperator__(binaryexpr)
        self.stack.append(self.combine([left, right], operator="<"))

    def visit_GTEQ(self, binaryexpr):
        left, right = self.__visit_BinaryOperator__(binaryexpr)
        self.stack.append(self.combine([left, right], operator=">="))

    def visit_LTEQ(self, binaryexpr):
        left, right = self.__visit_BinaryOperator__(binaryexpr)
        self.stack.append(self.combine([left, right], operator="<="))

    def visit_NamedAttributeRef(self, named):
        self.stack.append(self.language.compile_attribute(named))

    def visit_UnnamedAttributeRef(self, unnamed):
        LOG.debug("expr %s is UnnamedAttributeRef", unnamed)
        self.stack.append(self.language.compile_attribute(unnamed))

    def visit_NumericLiteral(self, numericliteral):
        self.stack.append(self.language.compile_numericliteral(numericliteral))

    def visit_StringLiteral(self, stringliteral):
        self.stack.append(self.language.compile_stringliteral(stringliteral))

    def visit_DIVIDE(self, binaryexpr):
        left, right = self.__visit_BinaryOperator__(binaryexpr)
        self.stack.append(self.combine([left, right], operator="/"))

    def visit_PLUS(self, binaryexpr):
        left, right = self.__visit_BinaryOperator__(binaryexpr)
        self.stack.append(self.combine([left, right], operator="+"))

    def visit_MINUS(self, binaryexpr):
        left, right = self.__visit_BinaryOperator__(binaryexpr)
        self.stack.append(self.combine([left, right], operator="-"))

    def visit_IDIVIDE(self, binaryexpr):
        left, right = self.__visit_BinaryOperator__(binaryexpr)
        self.stack.append(self.combine([left, right], operator="/"))

    def visit_TIMES(self, binaryexpr):
        left, right = self.__visit_BinaryOperator__(binaryexpr)
        self.stack.append(self.combine([left, right], operator="*"))

    def visit_NEG(self, unaryexpr):
        inputexpr = self.stack.pop()
        self.stack.append(self.language.negative(inputexpr))


import unittest

import raco.fakedb
from raco.relation_key import RelationKey
from raco.algebra import *
from raco.expression import *
import raco.relation_key as relation_key
from raco.expression import StateVar


class TestQueryFunctions():

    emp_table = collections.Counter([
        # id dept_id name salary
        (1, 2, "Bill Howe", 25000),
        (2, 1, "Dan Halperin", 90000),
        (3, 1, "Andrew Whitaker", 5000),
        (4, 2, "Shumo Chu", 5000),
        (5, 1, "Victor Almeida", 25000),
        (6, 3, "Dan Suciu", 90000),
        (7, 1, "Magdalena Balazinska", 25000)])

    emp_schema = scheme.Scheme([("id", types.LONG_TYPE),
                                ("dept_id", types.LONG_TYPE),
                                ("name", types.STRING_TYPE),
                                ("salary", types.LONG_TYPE)])

    emp_key = relation_key.RelationKey.from_string("andrew:adhoc:employee")


class OperatorTest(unittest.TestCase):
    def setUp(self):
        self.db = raco.fakedb.FakeDatabase()
        self.db.ingest(TestQueryFunctions.emp_key,
                       TestQueryFunctions.emp_table,
                       TestQueryFunctions.emp_schema)

    def test_counter_stateful_apply(self):
        """Test stateful apply operator that produces a counter"""
        scan = Scan(TestQueryFunctions.emp_key, TestQueryFunctions.emp_schema)

        initex = NumericLiteral(-1)
        iterex = NamedStateAttributeRef("count")
        updateex = PLUS(UnnamedStateAttributeRef(0),
                        NumericLiteral(1))

        sapply = StatefulApply([("count", iterex)],
                               [StateVar("count", initex, updateex)], scan)
        result = collections.Counter(self.db.evaluate(sapply))
        self.assertEqual(len(result), len(TestQueryFunctions.emp_table))
        self.assertEqual([x[0] for x in result], range(7))

    def test_times_equal_uda(self):
        input_op = Scan(TestQueryFunctions.emp_key,
                        TestQueryFunctions.emp_schema)

        init_ex = NumericLiteral(1)
        update_ex = TIMES(NamedStateAttributeRef("value"),
                          NamedAttributeRef("salary"))
        emit_ex = UdaAggregateExpression(NamedStateAttributeRef("value"))

        statemods = [StateVar("value", init_ex, update_ex)]
        gb = GroupBy([UnnamedAttributeRef(1)], [emit_ex], input_op, statemods)
        result = self.db.evaluate_to_bag(gb)

        d = collections.defaultdict(lambda: 1)
        for tpl in TestQueryFunctions.emp_table:
            d[tpl[1]] *= tpl[3]
        expected = collections.Counter(
            [(key, val) for key, val in d.iteritems()])

        self.assertEquals(result, expected)

    def test_running_mean_stateful_apply(self):
        """Calculate the mean using stateful apply"""
        scan = Scan(TestQueryFunctions.emp_key, TestQueryFunctions.emp_schema)

        initex0 = NumericLiteral(0)
        updateex0 = PLUS(NamedStateAttributeRef("count"),
                         NumericLiteral(1))

        initex1 = NumericLiteral(0)
        updateex1 = PLUS(NamedStateAttributeRef("sum"),
                         NamedAttributeRef("salary"))

        avgex = IDIVIDE(NamedStateAttributeRef("sum"),
                        NamedStateAttributeRef("count"))

        sapply = StatefulApply([("avg", avgex)],
                               [StateVar("count", initex0, updateex0),
                                StateVar("sum", initex1, updateex1)], scan)

        store = Store(RelationKey("OUTPUT"), sapply)
        result = list(self.db.evaluate(sapply))

        self.assertEqual(len(result), len(TestQueryFunctions.emp_table))
        self.assertEqual([x[0] for x in result][-1], 37857)

        # test whether we can generate json without errors
        from raco.language.myrialang import (compile_to_json,
                                             MyriaLeftDeepTreeAlgebra)
        from compile import optimize
        import json
        json_string = json.dumps(compile_to_json("", None, optimize(store, MyriaLeftDeepTreeAlgebra())))  # noqa
        assert json_string

    def test_cast_to_float(self):
        scan = Scan(TestQueryFunctions.emp_key, TestQueryFunctions.emp_schema)
        cast = CAST(types.DOUBLE_TYPE, NamedAttributeRef("salary"))
        applyop = Apply([("salaryf", cast)], scan)
        res = list(self.db.evaluate(applyop))
        for x in res:
            assert isinstance(x[0], float)
        self.assertEqual([x[0] for x in res],
                         [x[3] for x in TestQueryFunctions.emp_table])

    def test_projecting_join_scheme(self):
        emp = Scan(TestQueryFunctions.emp_key, TestQueryFunctions.emp_schema)
        emp1 = Scan(TestQueryFunctions.emp_key, TestQueryFunctions.emp_schema)
        pj = ProjectingJoin(condition=BooleanLiteral(True),
                            left=emp, right=emp1)
        names = ([n for n in emp.scheme().get_names()]
                 + ["{n}1".format(n=n) for n in emp.scheme().get_names()])
        self.assertEquals(names, pj.scheme().get_names())

    def test_projecting_join_scheme_no_dups_alternate(self):
        emp = Scan(TestQueryFunctions.emp_key, TestQueryFunctions.emp_schema)
        emp1 = Scan(TestQueryFunctions.emp_key, TestQueryFunctions.emp_schema)
        num_cols = len(emp.scheme())
        # alternate which copy of emp we keep a col from
        refs = [UnnamedAttributeRef(i + (i % 2) * num_cols)
                for i in range(num_cols)]
        pj = ProjectingJoin(condition=BooleanLiteral(True),
                            left=emp, right=emp1, output_columns=refs)
        self.assertEquals(emp.scheme().get_names(), pj.scheme().get_names())

    def test_projecting_join_scheme_no_dups_only_keep_right(self):
        emp = Scan(TestQueryFunctions.emp_key, TestQueryFunctions.emp_schema)
        emp1 = Scan(TestQueryFunctions.emp_key, TestQueryFunctions.emp_schema)
        num_cols = len(emp.scheme())
        # keep only the right child's columns
        refs = [UnnamedAttributeRef(i + num_cols)
                for i in range(num_cols)]
        pj = ProjectingJoin(condition=BooleanLiteral(True),
                            left=emp, right=emp1, output_columns=refs)
        self.assertEquals(emp.scheme().get_names(), pj.scheme().get_names())



import collections
import sys

from ply import yacc

from raco import relation_key
import raco.myrial.scanner as scanner
import raco.scheme as scheme
import raco.types
import raco.expression as sexpr
import raco.myrial.emitarg as emitarg
from raco.expression.udf import Function, StatefulFunc
import raco.expression.expressions_library as expr_lib
from .exceptions import *
import raco.types
from raco.expression import StateVar


class JoinColumnCountMismatchException(Exception):
    pass

# ID is a symbol name that identifies an input expression; columns is a list of
# columns expressed as either names or integer positions.
JoinTarget = collections.namedtuple('JoinTarget', ['expr', 'columns'])

SelectFromWhere = collections.namedtuple(
    'SelectFromWhere', ['distinct', 'select', 'from_', 'where', 'limit'])

DecomposableAgg = collections.namedtuple(
    'DecomposableAgg', ['logical', 'local', 'remote'])

# Mapping from source symbols to raco.expression.BinaryOperator classes
binops = {
    '+': sexpr.PLUS,
    '-': sexpr.MINUS,
    '/': sexpr.DIVIDE,
    '//': sexpr.IDIVIDE,
    '*': sexpr.TIMES,
    '>': sexpr.GT,
    '<': sexpr.LT,
    '>=': sexpr.GTEQ,
    '<=': sexpr.LTEQ,
    '!=': sexpr.NEQ,
    '<>': sexpr.NEQ,
    '==': sexpr.EQ,
    '=': sexpr.EQ,
    'AND': sexpr.AND,
    'OR': sexpr.OR,
}

# Map from myrial token name to raco internal type name.
myrial_type_map = {
    "STRING": raco.types.STRING_TYPE,
    "INT": raco.types.LONG_TYPE,
    "FLOAT": raco.types.DOUBLE_TYPE,
    "BOOLEAN": raco.types.BOOLEAN_TYPE
}


def contains_tuple_expression(ex):
    """Return True if an Expression contains a TupleExpression"""
    return any(isinstance(sx, TupleExpression) for sx in ex.walk())


def check_no_tuple_expression(ex, lineno):
    if contains_tuple_expression(ex):
        raise NestedTupleExpressionException(lineno)


def check_simple_expression(ex, lineno):
    check_no_tuple_expression(ex, lineno)
    sexpr.check_no_aggregate(ex, lineno)


def get_emitters(ex):
    if isinstance(ex, TupleExpression):
        return ex.emitters
    else:
        return [ex]


def get_num_emitters(ex):
    return len(get_emitters(ex))


class TupleExpression(sexpr.Expression):
    """Represents an instance of a tuple-valued Expression

    This class is a pseudo-expression that corresponds to a UDA or stateful
    apply with multiple return values.  Myria doesn't support tuples as a
    first-class data type.  Instead, instances of TupleExpression are converted
    into multiple scalar expression instances.
    """
    def __init__(self, emitters):
        self.emitters = emitters

    def walk(self):
        yield self
        for emitter in self.emitters:
            for x in emitter.walk():
                yield x

    def apply(self, f):
        self.emitters = [f(e) for e in self.emitters]

    def check_for_nested(self, lineno):
        """Raise an exception if a sub-expression contains a TupleExpression"""

        for ex in self.emitters:
            check_no_tuple_expression(ex, lineno)

    def get_children(self):
        return self.emitters

    def typeof(self, scheme, state_scheme):
        """Type checks are not applied to TupleExpressions."""
        raise NotImplementedError()

    def evaluate(self, _tuple, scheme, state=None):
        raise NotImplementedError()


class Parser(object):
    # mapping from function name to Function tuple
    udf_functions = {}

    # state modifier variables accessed by the current emit argument
    statemods = []

    # A unique ID pool for the stateful apply state variables
    mangle_id = 0

    # mapping from UDA name to local, remote aggregates
    decomposable_aggs = {}

    def __init__(self, log=yacc.PlyLogger(sys.stderr)):
        self.log = log
        self.tokens = scanner.tokens

        # Precedence among scalar expression operators in ascending order; this
        # is necessary to disambiguate the grammar.  Operator precedence is
        # identical to Python:
        # http://docs.python.org/2/reference/expressions.html#comparisons

        self.precedence = (
            ('left', 'OR'),
            ('left', 'AND'),
            ('right', 'NOT'),
            ('left', 'EQ', 'EQUALS', 'NE', 'GT', 'LT', 'LE', 'GE'),
            ('left', 'PLUS', 'MINUS'),
            ('left', 'TIMES', 'DIVIDE', 'IDIVIDE'),
            ('right', 'UMINUS'),    # Unary minus
        )

    # A MyriaL program consists of 1 or more "translation units", each of which
    # is a function, apply definition, or statement.
    @staticmethod
    def p_translation_unit_list(p):
        """translation_unit_list : translation_unit_list translation_unit
                                 | translation_unit"""
        if len(p) == 3:
            p[0] = p[1] + [p[2]]
        else:
            p[0] = [p[1]]

    @staticmethod
    def p_translation_unit(p):
        """translation_unit : statement
                            | constant
                            | udf
                            | apply
                            | uda
                            | decomposable_uda"""
        p[0] = p[1]

    @staticmethod
    def check_for_undefined(p, name, _sexpr, args):
        undefined = sexpr.udf_undefined_vars(_sexpr, args)
        if undefined:
            raise UndefinedVariableException(name, undefined[0], p.lineno(0))

    @staticmethod
    def check_for_reserved(p, name):
        """Check whether an identifier name is reserved."""
        if expr_lib.is_defined(name):
            raise ReservedTokenException(name, p.lineno(0))

    @staticmethod
    def add_decomposable_uda(p, logical, local, remote):
        """Register a decomposable UDA.

        :param p: The parser context
        :param logical: The name of the logical UDA
        :param local: The name of the local UDA
        :param remote: The name of the remote UDA
        """
        lineno = p.lineno(0)

        if logical in Parser.decomposable_aggs:
            raise DuplicateFunctionDefinitionException(logical, lineno)

        def check_name(name):
            if name not in Parser.udf_functions:
                raise NoSuchFunctionException(lineno)

            func = Parser.udf_functions[name]
            if not isinstance(func, StatefulFunc):
                raise NoSuchFunctionException(lineno)
            if not sexpr.expression_contains_aggregate(func.sexpr):
                raise NoSuchFunctionException(lineno)
            return func

        da = DecomposableAgg(*[check_name(x) for x in
                             (logical, local, remote)])

        # Do some basic sanity checking of the arguments; we can't do full
        # type inspection here, because the full type information is not
        # known until the function is invoked.

        # Number of local inputs must match number of logical inputs
        if len(da.local.args) != len(da.logical.args):
            raise InvalidArgumentList(local, da.logical.args, lineno)

        # Number of local outputs must equal number of remote inputs
        num_local_emitters = get_num_emitters(da.local.sexpr)
        if num_local_emitters != len(da.remote.args):
            phony_names = ['x%d' % n for n in range(num_local_emitters)]
            raise InvalidArgumentList(remote, phony_names, lineno)

        # Number of remote outputs must match number of logical outputs
        if get_num_emitters(da.logical.sexpr) != get_num_emitters(da.remote.sexpr):  # noqa
            raise InvalidEmitList(remote, lineno)

        Parser.decomposable_aggs[logical] = da

    @staticmethod
    def add_nary_udf(p, name, args, emitters):
        """Add an n-ary user-defined function to the global function table.

        :param p: The parser context
        :param name: The name of the function
        :type name: string
        :param args: A list of function arguments
        :type args: list of strings
        :param emitter: The output expression(s)
        :type body_expr: A list of NaryEmitArg instances
        """
        if not all(isinstance(e, emitarg.NaryEmitArg) for e in emitters):
            raise IllegalWildcardException(name, p.lineno(0))
        if sum(len(x.sexprs) for x in emitters) != len(emitters):
            raise NestedTupleExpressionException(p.lineno(0))
        emit_exprs = [e.sexprs[0] for e in emitters]
        Parser.add_udf(p, name, args, emit_exprs)

    @staticmethod
    def add_udf(p, name, args, body_exprs):
        """Add a user-defined function to the global function table.

        :param p: The parser context
        :param name: The name of the function
        :type name: string
        :param args: A list of function arguments
        :type args: list of strings
        :param body_exprs: A list of scalar expressions containing the body
        :type body_exprs: list of raco.expression.Expression
        """
        if name in Parser.udf_functions:
            raise DuplicateFunctionDefinitionException(name, p.lineno(0))

        if len(args) != len(set(args)):
            raise DuplicateVariableException(name, p.lineno(0))

        if len(body_exprs) == 1:
            emit_op = body_exprs[0]
        else:
            emit_op = TupleExpression(body_exprs)

        Parser.check_for_undefined(p, name, emit_op, args)

        Parser.udf_functions[name] = Function(args, emit_op)
        return emit_op

    @staticmethod
    def mangle(name):
        Parser.mangle_id += 1
        return "{name}__{mid}".format(name=name, mid=Parser.mangle_id)

    @staticmethod
    def add_state_func(p, name, args, inits, updates, emitters, is_aggregate):
        """Register a stateful apply or UDA.

        :param name: The name of the function
        :param args: A list of function argument names (strings)
        :param inits: A list of NaryEmitArg that describe init logic; each
        should contain exactly one emit expression.
        :param updates: A list of Expression that describe update logic
        :param emitters: An Expression list that returns the final results.
        If None, all statemod variables are returned in the order specified.
        :param is_aggregate: True if the state_func is a UDA

        TODO: de-duplicate logic from add_udf.
        """
        lineno = p.lineno(0)
        if name in Parser.udf_functions:
            raise DuplicateFunctionDefinitionException(name, lineno)
        if len(args) != len(set(args)):
            raise DuplicateVariableException(name, lineno)
        if len(inits) != len(updates):
            raise BadApplyDefinitionException(name, lineno)

        # Unpack the update, init expressions into a statemod dictionary
        statemods = collections.OrderedDict()
        for init, update in zip(inits, updates):
            if not isinstance(init, emitarg.NaryEmitArg):
                raise IllegalWildcardException(name, lineno)

            if len(init.sexprs) != 1:
                raise NestedTupleExpressionException(lineno)

            # Init, update expressions contain tuples or contain aggregates
            check_simple_expression(init.sexprs[0], lineno)
            check_simple_expression(update, lineno)

            if not init.column_names:
                raise UnnamedStateVariableException(name, lineno)

            # check for duplicate variable definitions
            sm_name = init.column_names[0]
            if sm_name in statemods or sm_name in args:
                raise DuplicateVariableException(name, lineno)

            statemods[sm_name] = (init.sexprs[0], update)

        # Check for undefined variables:
        #  - Init expressions cannot reference any variables.
        #  - Update expression can reference function arguments and state
        #    variables.
        #  - The emitter expressions can reference state variables.
        allvars = statemods.keys() + args
        for init_expr, update_expr in statemods.itervalues():
            Parser.check_for_undefined(p, name, init_expr, [])
            Parser.check_for_undefined(p, name, update_expr, allvars)

        if emitters is None:
            emitters = [sexpr.NamedAttributeRef(v) for v in statemods.keys()]

        for e in emitters:
            Parser.check_for_undefined(p, name, e, statemods.keys())
            check_simple_expression(e, lineno)

        # If the function is a UDA, wrap the output expression(s) so
        # downstream users can distinguish stateful apply from
        # aggregate expressions.
        if is_aggregate:
            emitters = [sexpr.UdaAggregateExpression(e) for e in emitters]

        assert len(emitters) > 0
        if len(emitters) == 1:
            emit_op = emitters[0]
        else:
            emit_op = TupleExpression(emitters)

        Parser.udf_functions[name] = StatefulFunc(args, statemods, emit_op)

    @staticmethod
    def p_unreserved_id(p):
        'unreserved_id : ID'
        Parser.check_for_reserved(p, p[1])
        p[0] = p[1]

    @staticmethod
    def p_unreserved_id_list(p):
        """unreserved_id_list : unreserved_id_list COMMA unreserved_id
                              | unreserved_id"""
        if len(p) == 4:
            p[0] = p[1] + [p[3]]
        else:
            p[0] = [p[1]]

    @staticmethod
    def p_udf(p):
        """udf : DEF unreserved_id LPAREN optional_arg_list RPAREN COLON sexpr SEMI"""  # noqa
        Parser.add_udf(p, p[2], p[4], [p[7]])
        p[0] = None

    @staticmethod
    def p_nary_udf(p):
        """udf : DEF unreserved_id LPAREN optional_arg_list RPAREN COLON table_literal SEMI"""  # noqa
        Parser.add_nary_udf(p, p[2], p[4], p[7])
        p[0] = None

    @staticmethod
    def p_constant(p):
        """constant : CONST unreserved_id COLON sexpr SEMI"""
        Parser.add_udf(p, p[2], [], [p[4]])
        p[0] = None

    @staticmethod
    def p_optional_arg_list(p):
        """optional_arg_list : function_arg_list
                             | empty"""
        p[0] = p[1] or []

    @staticmethod
    def p_function_arg_list(p):
        """function_arg_list : function_arg_list COMMA unreserved_id
                             | unreserved_id"""
        if len(p) == 4:
            p[0] = p[1] + [p[3]]
        else:
            p[0] = [p[1]]

    @staticmethod
    def p_statefunc_emit_list(p):
        """statefunc_emit_list : LBRACKET sexpr_list RBRACKET SEMI
                               | sexpr SEMI
                               | empty"""
        if len(p) == 5:
            p[0] = p[2]
        elif len(p) == 3:
            p[0] = (p[1],)
        else:
            p[0] = None

    @staticmethod
    def p_decomposable_uda(p):
        'decomposable_uda : UDA TIMES unreserved_id LBRACE unreserved_id COMMA unreserved_id RBRACE SEMI'  # noqa
        logical = p[3]
        local = p[5]
        remote = p[7]
        Parser.add_decomposable_uda(p, logical, local, remote)
        p[0] = None

    @staticmethod
    def p_uda(p):
        'uda : UDA unreserved_id LPAREN optional_arg_list RPAREN LBRACE \
        table_literal SEMI LBRACKET sexpr_list RBRACKET SEMI statefunc_emit_list RBRACE SEMI'  # noqa

        name = p[2]
        args = p[4]
        inits = p[7]
        updates = p[10]
        emits = p[13]
        Parser.add_state_func(p, name, args, inits, updates, emits, True)
        p[0] = None

    @staticmethod
    def p_apply(p):
        'apply : APPLY unreserved_id LPAREN optional_arg_list RPAREN LBRACE \
        table_literal SEMI LBRACKET sexpr_list RBRACKET SEMI statefunc_emit_list RBRACE SEMI'  # noqa

        name = p[2]
        args = p[4]
        inits = p[7]
        updates = p[10]
        emits = p[13]
        Parser.add_state_func(p, name, args, inits, updates, emits, False)
        p[0] = None

    @staticmethod
    def p_statement_assign(p):
        'statement : unreserved_id EQUALS rvalue SEMI'
        p[0] = ('ASSIGN', p[1], p[3])

    @staticmethod
    def p_statement_empty(p):
        'statement : SEMI'
        p[0] = None  # stripped out by parse

    # expressions must be embeddable in other expressions; certain constructs
    # are not embeddable, but are available as r-values in an assignment
    @staticmethod
    def p_rvalue(p):
        """rvalue : expression
                  | select_from_where"""
        p[0] = p[1]

    @staticmethod
    def p_statement_list(p):
        """statement_list : statement_list statement
                          | statement"""
        if len(p) == 3:
            p[0] = p[1] + [p[2]]
        else:
            p[0] = [p[1]]

    @staticmethod
    def p_statement_dowhile(p):
        'statement : DO statement_list WHILE expression SEMI'
        p[0] = ('DOWHILE', p[2], p[4])

    @staticmethod
    def p_statement_store(p):
        'statement : STORE LPAREN unreserved_id COMMA relation_key optional_part_info RPAREN SEMI'  # noqa
        p[0] = ('STORE', p[3], p[5], p[6])

    @staticmethod
    def p_statement_dump(p):
        'statement : DUMP LPAREN unreserved_id RPAREN SEMI'
        p[0] = ('DUMP', p[3])

    @staticmethod
    def p_optional_part_info(p):
        """optional_part_info : COMMA LBRACKET column_ref_list RBRACKET
                              | empty"""
        if len(p) > 2:
            p[0] = p[3]
        else:
            p[0] = None

    @staticmethod
    def p_expression_id(p):
        'expression : unreserved_id'
        p[0] = ('ALIAS', p[1])

    @staticmethod
    def p_sexpr_list(p):
        """sexpr_list : sexpr_list COMMA sexpr
                      | sexpr"""
        if len(p) == 4:
            p[0] = p[1] + (p[3],)
        else:
            p[0] = (p[1],)

    @staticmethod
    def p_expression_table_literal(p):
        'expression : table_literal'
        p[0] = ('TABLE', p[1])

    @staticmethod
    def p_table_literal(p):
        'table_literal : LBRACKET emit_arg_list RBRACKET'
        p[0] = p[2]

    @staticmethod
    def p_expression_empty(p):
        'expression : EMPTY LPAREN column_def_list RPAREN'
        p[0] = ('EMPTY', scheme.Scheme(p[3]))

    @staticmethod
    def p_expression_scan(p):
        'expression : SCAN LPAREN relation_key RPAREN'
        p[0] = ('SCAN', p[3])

    @staticmethod
    def p_expression_load(p):
        'expression :  LOAD LPAREN STRING_LITERAL COMMA column_def_list RPAREN'
        p[0] = ('LOAD', p[3], scheme.Scheme(p[5]))

    @staticmethod
    def p_relation_key(p):
        """relation_key : string_arg
                        | string_arg COLON string_arg
                        | string_arg COLON string_arg COLON string_arg"""
        p[0] = relation_key.RelationKey.from_string(''.join(p[1:]))

    # Note: column list cannot be empty
    @staticmethod
    def p_column_def_list(p):
        """column_def_list : column_def_list COMMA column_def
                           | column_def"""
        if len(p) == 4:
            cols = p[1] + [p[3]]
        else:
            cols = [p[1]]
        p[0] = cols

    @staticmethod
    def p_column_def(p):
        'column_def : unreserved_id COLON type_name'
        p[0] = (p[1], p[3])

    @staticmethod
    def p_type_name(p):
        """type_name : STRING
                     | INT
                     | BOOLEAN
                     | FLOAT"""
        p[0] = myrial_type_map[p[1]]

    @staticmethod
    def p_string_arg(p):
        """string_arg : unreserved_id
                      | STRING_LITERAL"""
        p[0] = p[1]

    @staticmethod
    def p_expression_bagcomp(p):
        'expression : LBRACKET FROM from_arg_list opt_where_clause \
        EMIT emit_arg_list RBRACKET'
        p[0] = ('BAGCOMP', p[3], p[4], p[6])

    @staticmethod
    def p_from_arg_list(p):
        """from_arg_list : from_arg_list COMMA from_arg
                         | from_arg"""
        if len(p) == 4:
            p[0] = p[1] + [p[3]]
        else:
            p[0] = [p[1]]

    @staticmethod
    def p_from_arg(p):
        """from_arg : expression optional_as unreserved_id
                    | unreserved_id"""
        expr = None
        if len(p) == 4:
            expr = p[1]
            _id = p[3]
        else:
            _id = p[1]
        p[0] = (_id, expr)

    @staticmethod
    def p_optional_as(p):
        """optional_as : AS
                       | empty"""
        p[0] = None

    @staticmethod
    def p_opt_where_clause(p):
        """opt_where_clause : WHERE sexpr
                            | empty"""
        if len(p) == 3:
            p[0] = p[2]
        else:
            p[0] = None

    @staticmethod
    def p_emit_arg_list(p):
        """emit_arg_list : emit_arg_list COMMA emit_arg
                         | emit_arg"""
        if len(p) == 4:
            p[0] = p[1] + (p[3],)
        else:
            p[0] = (p[1],)

    @staticmethod
    def p_emit_arg_explicit(p):
        """emit_arg : sexpr AS LBRACKET unreserved_id_list RBRACKET
                    | sexpr AS unreserved_id
                    | sexpr"""

        sx = p[1]
        names = None
        if len(p) == 6:
            names = p[4]
        if len(p) == 4:
            names = [p[3]]

        emitters = get_emitters(sx)
        if names is not None and len(emitters) != len(names):
            raise IllegalColumnNamesException(p.lineno(0))

        # Verify that there are no nested aggregate expressions
        for ssx in emitters:
            sexpr.check_no_nested_aggregate(ssx, p.lineno(0))

        # Verify that there are no remaining tuple expressions
        for ssx in emitters:
            check_no_tuple_expression(ssx, p.lineno(0))

        p[0] = emitarg.NaryEmitArg(names, emitters, Parser.statemods)
        Parser.statemods = []

    @staticmethod
    def p_emit_arg_table_wildcard(p):
        """emit_arg : unreserved_id DOT TIMES"""
        p[0] = emitarg.TableWildcardEmitArg(p[1])

    @staticmethod
    def p_emit_arg_full_wildcard(p):
        """emit_arg : TIMES"""
        p[0] = emitarg.FullWildcardEmitArg()

    @staticmethod
    def p_expression_select_from_where(p):
        """expression : LPAREN select_from_where RPAREN"""
        p[0] = p[2]

    @staticmethod
    def p_select_from_where(p):
        'select_from_where : SELECT opt_distinct emit_arg_list FROM from_arg_list opt_where_clause opt_limit'  # noqa
        p[0] = ('SELECT', SelectFromWhere(distinct=p[2], select=p[3],
                                          from_=p[5], where=p[6], limit=p[7]))

    @staticmethod
    def p_opt_distinct(p):
        """opt_distinct : DISTINCT
                        | empty"""
        # p[1] is either 'DISTINCT' or None. Use Python truthiness
        p[0] = bool(p[1])

    @staticmethod
    def p_opt_limit(p):
        """opt_limit : LIMIT INTEGER_LITERAL
                     | empty"""
        if len(p) == 3:
            p[0] = p[2]
        else:
            p[0] = None

    @staticmethod
    def p_expression_limit(p):
        'expression : LIMIT LPAREN expression COMMA INTEGER_LITERAL RPAREN'
        p[0] = ('LIMIT', p[3], p[5])

    @staticmethod
    def p_expression_distinct(p):
        'expression : DISTINCT LPAREN expression RPAREN'
        p[0] = ('DISTINCT', p[3])

    @staticmethod
    def p_expression_countall(p):
        'expression : COUNTALL LPAREN expression RPAREN'
        p[0] = ('COUNTALL', p[3])

    @staticmethod
    def p_expression_binary_set_operation(p):
        'expression : setop LPAREN expression COMMA expression RPAREN'
        p[0] = (p[1], p[3], p[5])

    @staticmethod
    def p_setop(p):
        """setop : INTERSECT
                 | DIFF
                 | UNIONALL"""
        p[0] = p[1]

    @staticmethod
    def p_expression_unionall_inline(p):
        """expression : expression PLUS expression"""
        p[0] = ('UNIONALL', p[1], p[3])

    @staticmethod
    def p_expression_cross(p):
        'expression : CROSS LPAREN expression COMMA expression RPAREN'
        p[0] = ('CROSS', p[3], p[5])

    @staticmethod
    def p_expression_join(p):
        'expression : JOIN LPAREN join_argument COMMA join_argument RPAREN'
        if len(p[3].columns) != len(p[5].columns):
            raise JoinColumnCountMismatchException()
        p[0] = ('JOIN', p[3], p[5])

    @staticmethod
    def p_join_argument_list(p):
        'join_argument : expression COMMA LPAREN column_ref_list RPAREN'
        p[0] = JoinTarget(p[1], p[4])

    @staticmethod
    def p_join_argument_single(p):
        'join_argument : expression COMMA column_ref'
        p[0] = JoinTarget(p[1], [p[3]])

    # column_ref refers to the name or position of a column; these serve
    # as arguments to join.
    @staticmethod
    def p_column_ref_list(p):
        """column_ref_list : column_ref_list COMMA column_ref
                           | column_ref"""
        if len(p) == 4:
            p[0] = p[1] + [p[3]]
        else:
            p[0] = [p[1]]

    @staticmethod
    def p_column_ref_string(p):
        'column_ref : unreserved_id'
        p[0] = p[1]

    @staticmethod
    def p_column_ref_index(p):
        'column_ref : DOLLAR INTEGER_LITERAL'
        p[0] = p[2]

    # scalar expressions map to raco.Expression instances; these are operations
    # that return scalar types.

    @staticmethod
    def p_sexpr_integer_literal(p):
        'sexpr : INTEGER_LITERAL'
        p[0] = sexpr.NumericLiteral(p[1])

    @staticmethod
    def p_sexpr_string_literal(p):
        'sexpr : STRING_LITERAL'
        p[0] = sexpr.StringLiteral(p[1])

    @staticmethod
    def p_sexpr_float_literal(p):
        'sexpr : FLOAT_LITERAL'
        p[0] = sexpr.NumericLiteral(p[1])

    @staticmethod
    def p_sexpr_boolean_literal(p):
        '''sexpr : TRUE
                 | FALSE'''
        bv = p[1] == 'TRUE'
        p[0] = sexpr.BooleanLiteral(bv)

    @staticmethod
    def p_sexpr_id(p):
        'sexpr : unreserved_id'
        try:
            # Check for zero-argument function
            p[0] = Parser.resolve_function(p, p[1], [])
        except:
            # Resolve as an attribute reference
            p[0] = sexpr.NamedAttributeRef(p[1])

    @staticmethod
    def p_sexpr_index(p):
        'sexpr : DOLLAR INTEGER_LITERAL'
        p[0] = sexpr.UnnamedAttributeRef(p[2])

    @staticmethod
    def p_sexpr_id_dot_id(p):
        'sexpr : unreserved_id DOT unreserved_id'
        p[0] = sexpr.Unbox(p[1], p[3])

    @staticmethod
    def p_sexpr_id_dot_pos(p):
        'sexpr : unreserved_id DOT DOLLAR INTEGER_LITERAL'
        p[0] = sexpr.Unbox(p[1], p[4])

    @staticmethod
    def p_sexpr_group(p):
        'sexpr : LPAREN sexpr RPAREN'
        p[0] = p[2]

    @staticmethod
    def p_sexpr_uminus(p):
        'sexpr : MINUS sexpr %prec UMINUS'
        p[0] = sexpr.TIMES(sexpr.NumericLiteral(-1), p[2])

    @staticmethod
    def p_sexpr_worker_id(p):
        """sexpr : WORKER_ID LPAREN RPAREN"""
        p[0] = sexpr.WORKERID()

    @staticmethod
    def p_sexpr_binop(p):
        """sexpr : sexpr PLUS sexpr
                   | sexpr MINUS sexpr
                   | sexpr TIMES sexpr
                   | sexpr DIVIDE sexpr
                   | sexpr IDIVIDE sexpr
                   | sexpr GT sexpr
                   | sexpr LT sexpr
                   | sexpr GE sexpr
                   | sexpr LE sexpr
                   | sexpr NE sexpr
                   | sexpr NE2 sexpr
                   | sexpr EQ sexpr
                   | sexpr EQUALS sexpr
                   | sexpr AND sexpr
                   | sexpr OR sexpr"""
        p[0] = binops[p[2]](p[1], p[3])

    @staticmethod
    def p_sexpr_not(p):
        'sexpr : NOT sexpr'
        p[0] = sexpr.NOT(p[2])

    @staticmethod
    def resolve_stateful_func(func, args):
        """Resolve a stateful function given argument expressions.

        :param func: An instance of StatefulFunc
        :param args: A list of argument expressions
        :return: An emit expression and a StateVar list.  All expressions
        have no free variables.
        """
        assert isinstance(func, StatefulFunc)
        state_var_names = func.statemods.keys()

        # Mangle state variable names to allow multiple invocations to coexist
        state_vars_mangled = [Parser.mangle(sv) for sv in state_var_names]
        mangle_dict = dict(zip(state_var_names, state_vars_mangled))

        statemods = []
        for name, (init_expr, update_expr) in func.statemods.iteritems():
            # Convert state mod references into appropriate expressions
            update_expr = sexpr.resolve_state_vars(update_expr,  # noqa
                state_var_names, mangle_dict)
            # Convert argument references into appropriate expressions
            update_expr = sexpr.resolve_function(update_expr,  # noqa
                dict(zip(func.args, args)))
            statemods.append(StateVar(mangle_dict[name],
                                      init_expr, update_expr))
        emit_expr = sexpr.resolve_state_vars(func.sexpr, state_var_names,
                                             mangle_dict)
        return emit_expr, statemods

    @staticmethod
    def resolve_function(p, name, args):
        """Resolve a function invocation into an Expression instance.

        :param p: The parser context
        :param name: The name of the function
        :type name: string
        :param args: A list of argument expressions
        :type args: list of raco.expression.Expression instances
        :return: An expression with no free variables.
        """

        # try to get function from udf or system defined functions
        if name in Parser.udf_functions:
            func = Parser.udf_functions[name]
        else:
            func = expr_lib.lookup(name, len(args))

        if func is None:
            raise NoSuchFunctionException(name, p.lineno(0))
        if len(func.args) != len(args):
            raise InvalidArgumentList(name, func.args, p.lineno(0))

        if isinstance(func, Function):
            return sexpr.resolve_function(func.sexpr, dict(zip(func.args, args)))  # noqa
        elif isinstance(func, StatefulFunc):
            emit_expr, statemods = Parser.resolve_stateful_func(func, args)
            Parser.statemods.extend(statemods)

            # If the aggregate is decomposable, construct local and remote
            # emitters and statemods.
            if name in Parser.decomposable_aggs:
                ds = Parser.decomposable_aggs[name]
                local_emit, local_statemods = Parser.resolve_stateful_func(
                    ds.local, args)

                # Problem: we must connect the local aggregate outputs to
                # the remote aggregate inputs.  At this stage, we don't have
                # enough information to construct argument expressions to
                # serve as input to the remote aggregate.  Instead, we
                # introduce a placeholder reference, which is referenced
                # relative to the start of the local aggregate output.
                remote_args = [sexpr.LocalAggregateOutput(i)
                               for i in range(len(ds.remote.args))]
                remote_emit, remote_statemods = Parser.resolve_stateful_func(
                    ds.remote, remote_args)

                # local and remote emitters may be tuple-valued; flatten them.
                local_emitters = get_emitters(local_emit)
                remote_emitters = get_emitters(remote_emit)
                ds = sexpr.DecomposableAggregateState(
                    local_emitters, local_statemods,
                    remote_emitters, remote_statemods)

                # Associate a decomposable state structure with the first
                # emitter.  Mark the remaining emitters as decomposable, but
                # without their own associated decomposed emitters and
                # statemods.
                emitters = get_emitters(emit_expr)
                emitters[0].set_decomposable_state(ds)
                for emt in emitters[1:]:
                    emt.set_decomposable_state(
                        sexpr.DecomposableAggregateState())
            return emit_expr
        else:
            assert False

    @staticmethod
    def p_sexpr_function_k_args(p):
        'sexpr : ID LPAREN function_param_list RPAREN'
        p[0] = Parser.resolve_function(p, p[1], p[3])

    @staticmethod
    def p_sexpr_function_zero_args(p):
        'sexpr : ID LPAREN RPAREN'
        p[0] = Parser.resolve_function(p, p[1], [])

    @staticmethod
    def p_function_param_list(p):
        """function_param_list : function_param_list COMMA sexpr
                               | sexpr"""
        if len(p) == 4:
            p[0] = p[1] + [p[3]]
        else:
            p[0] = [p[1]]

    @staticmethod
    def p_sexpr_countall(p):
        'sexpr : COUNTALL LPAREN RPAREN'
        p[0] = Parser.resolve_function(p, 'COUNTALL', [])

    @staticmethod
    def p_sexpr_count(p):
        'sexpr : COUNT LPAREN count_arg RPAREN'
        if p[3] == '*':
            p[0] = sexpr.COUNTALL()
        else:
            p[0] = sexpr.COUNT(p[3])

    @staticmethod
    def p_sexpr_cast(p):
        """sexpr : type_name LPAREN sexpr RPAREN"""
        p[0] = sexpr.CAST(p[1], p[3])

    @staticmethod
    def p_count_arg(p):
        """count_arg : TIMES
                     | sexpr"""
        p[0] = p[1]

    @staticmethod
    def p_sexpr_unbox(p):
        'sexpr : TIMES expression optional_column_ref'
        p[0] = sexpr.Unbox(p[2], p[3])

    @staticmethod
    def p_when_expr(p):
        'when_expr : WHEN sexpr THEN sexpr'
        p[0] = (p[2], p[4])

    @staticmethod
    def p_when_expr_list(p):
        """when_expr_list : when_expr_list when_expr
                          | when_expr
        """
        if len(p) == 3:
            p[0] = p[1] + [p[2]]
        else:
            p[0] = [p[1]]

    @staticmethod
    def p_sexpr_case(p):
        'sexpr : CASE when_expr_list ELSE sexpr END'
        p[0] = sexpr.Case(p[2], p[4])

    @staticmethod
    def p_optional_column_ref(p):
        """optional_column_ref : DOT column_ref
                               | empty"""
        if len(p) == 3:
            p[0] = p[2]
        else:
            p[0] = None

    @staticmethod
    def p_empty(p):
        'empty :'
        pass

    def parse(self, s):
        scanner.lexer.lineno = 1
        Parser.udf_functions = {}
        Parser.decomposable_aggs = {}
        parser = yacc.yacc(module=self, debug=False, optimize=False)
        stmts = parser.parse(s, lexer=scanner.lexer, tracking=True)

        # Strip out the remnants of parsed functions to leave only a list of
        # statements
        return [st for st in stmts if st is not None]

    @staticmethod
    def p_error(token):
        if token:
            raise MyrialParseException(token)
        else:
            raise MyrialUnexpectedEndOfFileException()



import collections

import raco.scheme as scheme
import raco.datalog.datalog_test as datalog_test
from raco import types
from raco.language.myrialang import MyriaHyperCubeAlgebra


class TestQueryFunctions(datalog_test.DatalogTestCase):
    emp_table = collections.Counter([
        # id dept_id name salary
        (1, 2, "Bill Howe", 25000),
        (2, 1, "Dan Halperin", 90000),
        (3, 1, "Andrew Whitaker", 5000),
        (4, 2, "Shumo Chu", 5000),
        (5, 1, "Victor Almeida", 25000),
        (6, 3, "Dan Suciu", 90000),
        (7, 1, "Magdalena Balazinska", 25000)])

    emp_schema = scheme.Scheme([("id", types.LONG_TYPE),
                                ("dept_id", types.LONG_TYPE),
                                ("name", types.STRING_TYPE),
                                ("salary", types.LONG_TYPE)])

    emp_key = "employee"

    dept_table = collections.Counter([
        (1, "accounting", 5),
        (2, "human resources", 2),
        (3, "engineering", 2),
        (4, "sales", 7)])

    dept_schema = scheme.Scheme([("id", types.LONG_TYPE),
                                 ("name", types.STRING_TYPE),
                                 ("manager", types.LONG_TYPE)])

    dept_key = "department"

    edge_table = collections.Counter([
        (1, 2),
        (2, 3),
        (3, 4),
        (4, 3),
        (3, 5),
        (4, 13),
        (5, 4),
        (1, 9),
        (7, 1),
        (6, 1),
        (10, 11),
        (11, 12),
        (12, 10),
        (13, 4),
        (10, 1)])

    edge_schema = scheme.Scheme([("src", types.LONG_TYPE),
                                 ("dst", types.LONG_TYPE)])
    edge_key = "Edge"

    def setUp(self):
        super(TestQueryFunctions, self).setUp()

        self.db.ingest(TestQueryFunctions.emp_key,
                       TestQueryFunctions.emp_table,
                       TestQueryFunctions.emp_schema)

        self.db.ingest(TestQueryFunctions.dept_key,
                       TestQueryFunctions.dept_table,
                       TestQueryFunctions.dept_schema)

        self.db.ingest(TestQueryFunctions.edge_key,
                       TestQueryFunctions.edge_table,
                       TestQueryFunctions.edge_schema)

    def test_simple_join(self):
        expected = collections.Counter(
            [(e[2], d[1]) for e in self.emp_table.elements()
             for d in self.dept_table.elements() if e[1] == d[0]])

        query = """
        EmpDepts(emp_name, dept_name) :- employee(a, dept_id, emp_name, b),
                department(dept_id, dept_name, c)
        """

        self.check_result(query, expected, output='EmpDepts')

    def test_filter(self):
        query = """
        RichGuys(name) :- employee(a, b, name, salary), salary > 25000
        """

        expected = collections.Counter(
            [(x[2],) for x in TestQueryFunctions.emp_table.elements()
             if x[3] > 25000])
        self.check_result(query, expected, output='RichGuys')

    def test_count(self):
        query = """
        OutDegree(src, count(dst)) :- Edge(src, dst)
        """

        counter = collections.Counter()
        for (src, _) in self.edge_table.elements():
            counter[src] += 1

        ex = [(src, cnt) for src, cnt in counter.iteritems()]
        expected = collections.Counter(ex)
        self.check_result(query, expected, output='OutDegree')

    def test_sum_reorder(self):
        query = """
        SalaryByDept(sum(salary), dept_id) :- employee(id, dept_id, name, salary);"""  # noqa
        results = collections.Counter()
        for emp in self.emp_table.elements():
            results[emp[1]] += emp[3]
        expected = collections.Counter([(y, x) for x, y in results.iteritems()])  # noqa
        self.check_result(query, expected, output='SalaryByDept')

    def test_aggregate_no_groups(self):
        query = """
        Total(count(x)) :- Edge(x, y)
        """
        expected = collections.Counter([
            (len(self.edge_table),)])
        self.check_result(query, expected, output='Total')

    def test_multiway_join_chained(self):
        query = """
        OneHop(x) :- Edge(1, x);
        TwoHop(x) :- OneHop(y), Edge(y, x);
        ThreeHop(x) :- TwoHop(y), Edge(y, x)
        """

        expected = collections.Counter([(4,), (5,)])
        self.check_result(query, expected, output='ThreeHop')

    def test_triangles(self):
        # TODO. Right now we have do this separately so that the x<y and y<z
        # conditions are not put in the Join, rather rendered as Selects.
        # Myrialang barfs on theta-joins.
        query = """
        T(x,y,z) :- Edge(x,y), Edge(y,z), Edge(z,x);
        A(x,y,z) :- T(x,y,z), x < y, x < z.
        """

        expected = collections.Counter([(3, 5, 4), (10, 11, 12)])
        self.check_result(query, expected, output='A')

    def test_multiway_join(self):
        query = """
        ThreeHop(z) :- Edge(1, x), Edge(x,y), Edge(y, z);
        """
        expected = collections.Counter([(4,), (5,)])
        self.check_result(query, expected, output='ThreeHop')

    def test_multiway_join_hyper_cube(self):
        query = """
        ThreeHop(z) :- Edge(1, x), Edge(x,y), Edge(y, z);
        """
        expected = collections.Counter([(4,), (5,)])
        self.check_result(query, expected, output='ThreeHop',
                          algebra=MyriaHyperCubeAlgebra)

    def test_union(self):
        query = """
        OUTPUT(b) :- {emp}(a, b, c, d)
        OUTPUT(b) :- {edge}(b, a)
        """.format(emp=self.emp_key, edge=self.edge_key)
        expected = collections.Counter(
            set([(b,) for (a, b, c, d) in self.emp_table]
                + [(b,) for (b, a) in self.edge_table])
        )
        self.check_result(query, expected, test_logical=True)

    def test_filter_expression(self):
        query = """
        OUTPUT(a, b, c) :- {emp}(a, b, c, d), d >= 25000, d < 91000
        """.format(emp=self.emp_key)
        expected = collections.Counter(
            [(a, b, c)
             for (a, b, c, d) in self.emp_table
             if (d >= 25000 and d < 91000)]
        )
        self.check_result(query, expected)

    def test_attributes_forward(self):
        """test that attributes are correct amid multiple conditions"""
        query = """
        OUTPUT(a) :- {edge}(a, b), {emp}(c, a, x, y), b=c
        """.format(emp=self.emp_key, edge=self.edge_key)
        expected = collections.Counter(
            [(a,)
             for (a, b) in self.edge_table
             for (c, a2, x, y) in self.emp_table
             if (a == a2 and b == c)]
        )
        self.check_result(query, expected)

    def test_attributes_reverse(self):
        """test that attributes are correct amid multiple conditions and when
        the order of variables in the terms is the opposite of the explicit
        condition"""
        query = """
        OUTPUT(a) :- {edge}(a, b), {emp}(c, a, x, y), c=b
        """.format(emp=self.emp_key, edge=self.edge_key)
        expected = collections.Counter(
            [(a,)
             for (a, b) in self.edge_table
             for (c, a2, x, y) in self.emp_table
             if (a == a2 and b == c)]
        )
        self.check_result(query, expected)

    def test_apply_head(self):
        query = """
        OUTPUT(a/b) :- {emp}(a, b, c, d)
        """.format(emp=self.emp_key)
        expected = collections.Counter([
            (a * 1.0 / b,) for (a, b, _, _) in self.emp_table
        ])
        self.check_result(query, expected)

    def test_aggregate_head(self):
        query = """
        OUTPUT(SUM(a)) :- {emp}(a, b, c, d)
        """.format(emp=self.emp_key)
        expected = collections.Counter([
            (sum(a for (a, _, _, _) in self.emp_table),)
        ])
        self.check_result(query, expected)

    def test_twoaggregate_head(self):
        query = """
        OUTPUT(SUM(a), COUNT(b)) :- {emp}(a, b, c, d)
        """.format(emp=self.emp_key)
        expected = collections.Counter([
            (sum(a for (a, _, _, _) in self.emp_table),
             sum(1 for (_, b, _, _) in self.emp_table))
        ])
        self.check_result(query, expected)

    def test_aggregate_head_group_self(self):
        query = """
        OUTPUT(SUM(a), b) :- {emp}(a, b, c, d)
        """.format(emp=self.emp_key)
        B = set(b for (_, b, _, _) in self.emp_table)
        expected = collections.Counter([
            (sum(a for (a, b, _, _) in self.emp_table
                 if b == b2), b2)
            for b2 in B
        ])
        self.check_result(query, expected)

    def test_aggregate_head_group_swap(self):
        query = """
        OUTPUT(b,SUM(a)) :- {emp}(a, b, c, d)
        """.format(emp=self.emp_key)
        B = set(b for (_, b, _, _) in self.emp_table)
        expected = collections.Counter([
            (b2, sum(a for (a, b, _, _) in self.emp_table
                     if b == b2))
            for b2 in B
        ])
        self.check_result(query, expected)

    def test_binop_aggregates(self):
        query = """
        OUTPUT(SUM(b)+SUM(a)) :- {emp}(a, b, c, d)
        """.format(emp=self.emp_key)
        expected = collections.Counter([
            (sum(b for (a, b, _, _) in self.emp_table) +
             sum(a for (a, b, _, _) in self.emp_table),)
        ])
        self.check_result(query, expected)

    def test_aggregate_of_binop(self):
        query = """
        OUTPUT(SUM(b+a)) :- {emp}(a, b, c, d)
        """.format(emp=self.emp_key)
        expected = collections.Counter(
            [(sum([(a + b) for (a, b, c, d) in self.emp_table]),)])
        self.check_result(query, expected)

    def test_literal_expr(self):
        query = """
        OUTPUT(z+1) :- Edge(z, y)
        """
        expected = collections.Counter([(z + 1,)
                                        for (z, _) in self.edge_table])
        self.check_result(query, expected)




import collections

import raco.scheme as scheme
import raco.myrial.myrial_test as myrial_test
from raco import types


class SigmaClippingTest(myrial_test.MyrialTestCase):
    points = [25.0, 27.2, 23.4, 25.1, 26.3, 24.9, 23.5, 22.7, 108.2,
              26.2, 25.3, 24.7, 25.01, 26.1, 22.8, 2.2, 24.8, 25.05, 25.15]
    points_tuples = [(i, x) for i, x in enumerate(points)]
    points_table = collections.Counter(points_tuples)

    points_schema = scheme.Scheme([('id', types.LONG_TYPE), ('v', types.DOUBLE_TYPE)])  # noqa
    points_key = "public:adhoc:sc_points"

    def setUp(self):
        super(SigmaClippingTest, self).setUp()

        self.db.ingest(SigmaClippingTest.points_key,
                       SigmaClippingTest.points_table,
                       SigmaClippingTest.points_schema)

        # TODO: Better support for empty relations in the language
        self.db.ingest("empty", collections.Counter(),
                       SigmaClippingTest.points_schema)

    def run_it(self, query):
        points = [(i, x) for i, x in self.points_tuples if x < 28 and x > 22]
        expected = collections.Counter(points)
        self.check_result(query, expected, output='sc_points_clipped')

    def test_v0(self):
        with open('examples/sigma-clipping-v0.myl') as fh:
            query = fh.read()
        self.run_it(query)

    def test_v2(self):
        with open('examples/sigma-clipping.myl') as fh:
            query = fh.read()
        self.run_it(query)



import collections


def emit(*args):
    """Return blocks of code as a string."""
    return "\n".join([str(x) for x in args if len(str(x)) > 0])


def emitlist(argslist):
    """Return blocks of code as a string."""
    return "\n".join([str(x) for x in argslist if len(str(x)) > 0])


def real_str(obj, skip_out=False):
    """Convert the input object to a string, recursively stringifying elements
    inside of containers. If skip_out is True, the container bounds will not
    be displayed. E.g. real_str([1, 2]) == "[1,2]" but
    real_str([1, 2], skip_out=True) == "1,2"."""

    # Hack around basestrings being containers
    if (not isinstance(obj, basestring)
            and isinstance(obj, collections.Container)):

        if isinstance(obj, collections.Sequence):
            inner = ','.join(real_str(e) for e in obj)
            if skip_out:
                return inner
            return '[{inn}]'.format(inn=inner)
        elif isinstance(obj, collections.Mapping):
            inner = ','.join('{a}:{b}'.format(a=real_str(a), b=real_str(b))
                             for a, b in obj.items())
            if skip_out:
                return inner
            return '{{{inn}}}'.format(inn=inner)
        elif isinstance(obj, collections.Set):
            inner = ','.join(real_str(e) for e in obj)
            if skip_out:
                return inner
            return '{{{inn}}}'.format(inn=inner)
        else:
            raise NotImplementedError(
                "real_str(obj) for type(obj)={t}".format(t=type(obj)))

    return str(obj)


class Printable(object):
    @classmethod
    def opname(cls):
        return str(cls.__name__)

    def __str__(self):
        return self.opname()


# Optional raco dependency: termcolor
# Without it, coloring will not happen
def colored(s, color):
    return s
try:
    from termcolor import colored
except ImportError:
    pass



import collections

# A user-defined function
Function = collections.namedtuple('Function', ['args', 'sexpr'])

# A user-defined stateful apply or UDA
StatefulFunc = collections.namedtuple(
    'StatefulFunc', ['args', 'statemods', "sexpr"])




class MyrialCompileException(Exception):
    pass


class MyrialUnexpectedEndOfFileException(MyrialCompileException):
    def __str__(self):
        return "Unexpected end-of-file"


class MyrialParseException(MyrialCompileException):
    def __init__(self, token):
        self.token = token

    def __str__(self):
        return 'Parse error at token %s on line %d' % (self.token.value,
                                                       self.token.lineno)


class MyrialScanException(MyrialCompileException):
    def __init__(self, token):
        self.token = token

    def __str__(self):
        return 'Illegal token string %s on line %d' % (self.token.value,
                                                       self.token.lineno)


class DuplicateFunctionDefinitionException(MyrialCompileException):
    def __init__(self, funcname, lineno):
        self.funcname = funcname
        self.lineno = lineno

    def __str__(self):
        return 'Duplicate function definition for %s on line %d' % (self.funcname,  # noqa
                                                                    self.lineno)  # noqa


class NoSuchFunctionException(MyrialCompileException):
    def __init__(self, funcname, lineno):
        self.funcname = funcname
        self.lineno = lineno

    def __str__(self):
        return 'No such function definition for %s on line %d' % (self.funcname,  # noqa
                                                                  self.lineno)  # noqa


class ReservedTokenException(MyrialCompileException):
    def __init__(self, token, lineno):
        self.token = token
        self.lineno = lineno

    def __str__(self):
        return 'The token "%s" on line %d is reserved.' % (self.token,
            self.lineno)  # noqa


class InvalidArgumentList(MyrialCompileException):
    def __init__(self, funcname, expected_args, lineno):
        self.funcname = funcname
        self.expected_args = expected_args
        self.lineno = lineno

    def __str__(self):
        return "Incorrect number of arguments for %s(%s) on line %d" % (
            self.funcname, ','.join(self.expected_args), self.lineno)


class UndefinedVariableException(MyrialCompileException):
    def __init__(self, funcname, var, lineno):
        self.funcname = funcname
        self.var = var
        self.lineno = lineno

    def __str__(self):
        return "Undefined variable %s in function %s at line %d" % (
            self.var, self.funcname, self.lineno)


class DuplicateVariableException(MyrialCompileException):
    def __init__(self, funcname, lineno):
        self.funcname = funcname
        self.lineno = lineno

    def __str__(self):
        return "Duplicately defined in function %s at line %d" % (
            self.funcname, self.lineno)


class BadApplyDefinitionException(MyrialCompileException):
    def __init__(self, funcname, lineno):
        self.funcname = funcname
        self.lineno = lineno

    def __str__(self):
        return "Bad apply definition for in function %s at line %d" % (
            self.funcname, self.lineno)


class UnnamedStateVariableException(MyrialCompileException):
    def __init__(self, funcname, lineno):
        self.funcname = funcname
        self.lineno = lineno

    def __str__(self):
        return "Unnamed state variable in function %s at line %d" % (
            self.funcname, self.lineno)


class IllegalWildcardException(MyrialCompileException):
    def __init__(self, funcname, lineno):
        self.funcname = funcname
        self.lineno = lineno

    def __str__(self):
        return "Illegal use of wildcard in function %s at line %d" % (
            self.funcname, self.lineno)


class NestedTupleExpressionException(MyrialCompileException):
    def __init__(self, lineno):
        self.lineno = lineno

    def __str__(self):
        return "Illegal use of tuple expression on line %d" % self.lineno


class InvalidEmitList(MyrialCompileException):
    def __init__(self, function, lineno):
        self.function = function
        self.lineno = lineno

    def __str__(self):
        return "Wrong number of emit arguments in %s at line %d" % (
            self.function, self.lineno)


class IllegalColumnNamesException(MyrialCompileException):
    def __init__(self, lineno):
        self.lineno = lineno

    def __str__(self):
        return "Invalid column names on line %d" % self.lineno


class ColumnIndexOutOfBounds(Exception):
    pass


class SchemaMismatchException(MyrialCompileException):
    def __init__(self, op_name):
        self.op_name = op_name

    def __str__(self):
        return "Incompatible input schemas for %s operation" % self.op_name



import collections
import unittest
import itertools

from raco.dbconn import DBConnection
from raco.fake_data import FakeData

"""Test the raco to sqllite interface."""


class SQLLiteTest(unittest.TestCase, FakeData):

    def setUp(self):
        self.conn1 = DBConnection()
        self.conn2 = DBConnection()

        self.conn1.add_table("emp", FakeData.emp_schema, FakeData.emp_table)
        self.conn1.add_table("dept", FakeData.dept_schema, FakeData.dept_table)
        self.conn2.add_table("num", FakeData.numbers_schema,
                             FakeData.numbers_table)

    def test_empty_relation(self):
        self.conn1.add_table("emp2", FakeData.emp_schema,
                             collections.Counter())
        emp_out = collections.Counter(self.conn1.get_table('emp2'))
        self.assertEquals(emp_out, collections.Counter())

        scheme_out = self.conn1.get_scheme('emp2')
        self.assertEquals(scheme_out, FakeData.emp_schema)

    def test_scan(self):
        emp_out = collections.Counter(self.conn1.get_table('emp'))
        self.assertEquals(emp_out, FakeData.emp_table)

        dept_out = collections.Counter(self.conn1.get_table('dept'))
        self.assertEquals(dept_out, FakeData.dept_table)

        num_out = collections.Counter(self.conn2.get_table('num'))
        self.assertEquals(num_out, FakeData.numbers_table)

    def test_schema_lookup(self):
        self.assertEquals(self.conn1.get_scheme('emp'), FakeData.emp_schema)
        self.assertEquals(self.conn1.get_scheme('dept'), FakeData.dept_schema)
        self.assertEquals(self.conn2.get_scheme('num'),
                          FakeData.numbers_schema)

    def test_num_tuples(self):
        self.assertEquals(self.conn1.num_tuples('emp'),
                          len(FakeData.emp_table))
        self.assertEquals(self.conn1.num_tuples('dept'),
                          len(FakeData.dept_table))
        self.assertEquals(self.conn2.num_tuples('num'),
                          len(FakeData.numbers_table))

    def test_schema_lookup_key_error(self):
        with self.assertRaises(KeyError):
            sc = self.conn2.get_scheme("emp")

    def test_scan_key_error(self):
        with self.assertRaises(KeyError):
            sc = self.conn1.get_table("num")

    def test_delete_table(self):
        sc = self.conn1.get_scheme("emp")
        self.conn1.delete_table("emp")
        with self.assertRaises(KeyError):
            sc = self.conn1.get_scheme("emp")

    def test_append_table(self):
        self.conn1.append_table("emp", FakeData.emp_table)

        it = itertools.chain(iter(FakeData.emp_table),
                             iter(FakeData.emp_table))
        expected = collections.Counter(it)
        actual = collections.Counter(self.conn1.get_table('emp'))
        self.assertEquals(actual, expected)
